from typing import Union, cast

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from colorspacious import cspace_converter  # type: ignore
from matplotlib.figure import Figure
from numpy.typing import NDArray

from arcadia_pycolor.gradient import Gradient
from arcadia_pycolor.gradients import all_gradients
from arcadia_pycolor.palettes import all_palettes


def plot_gradient_lightness(
    gradients: Union[Gradient, str, list[Gradient], list[str]],
    title: Union[str, None] = None,
    horizontal_spacing: float = 1.1,
    steps: int = 100,
    figsize: tuple[float, float] = (4, 4),
    cmap_type: str = "linear",
    tickrotation: float = 50,
    markersize: float = 300,
    return_fig: bool = False,
) -> Union[None, Figure]:
    """Plots the lightness of one or more color gradients to assess their uniformity.

    Args:
        gradients (list[Gradient] or list[str]):
            The list of gradients to plot, where each element is either a `Gradient` object
            or a string. If a string is provided, it is assumed to be a registered Matplotlib
            colormap.
        title (str):
            A title for the plot if desired.
        horizontal_spacing (float):
            The spacing between lines.
        steps (int):
            The number of steps along the gradient to generate.
        figsize (tuple):
            The width, height tuple of the figure size.
        cmap_type (str):
            'linear' if you want the label for the `cmap` to be at the end;
            anything else puts the label in the middle.
        tickrotation (int):
            The rotation of the label for each `cmap`.
        markersize (int):
            The size of the points that make up the gradient color line.
        return_fig (bool):
            Whether or not to return the figure as an object.

    Returns:
        Union[None, matplotlib.figure.Figure]:
            None if `return_fig` is False, otherwise a `matplotlib.figure.Figure` object.

    Raises:
        TypeError:
            If `gradients` is not a list of `Gradient` objects or strings.
    """

    fig, ax = plt.subplots(figsize=figsize, layout="constrained")  # type: ignore

    # Indices to step through colormap.
    x = np.linspace(0.0, 1.0, steps)

    xaxis_tick_labels: list[str] = []
    xaxis_tick_locations: list[float] = []

    # Check separately for single strings and single Gradient objects in order to avoid type errors.
    if isinstance(gradients, str):
        gradients = [gradients]
    elif isinstance(gradients, Gradient):
        gradients = [gradients]

    for ind, gradient in enumerate(gradients):
        if isinstance(gradient, str):
            gradient_name = gradient
            if gradient_name not in mpl.colormaps:  # type: ignore
                print(f"Colormap {gradient_name} not found in Matplotlib colormaps.")
                continue
            cmap = mpl.colormaps[gradient_name]
        elif isinstance(gradient, Gradient):
            gradient_name = gradient.name
            cmap = gradient.to_mpl_cmap()
        else:
            raise TypeError("gradients must be a list of Gradient objects or strings.")

        xaxis_tick_labels.append(gradient_name)
        cmap_as_rgb = cast(NDArray[np.float64], cmap(x)[np.newaxis, :, :3])
        cmap_as_lab = cast(NDArray[np.float64], cspace_converter("sRGB1", "CAM02-UCS")(cmap_as_rgb))

        # Plot colormap lightness values. Do this separately for each category
        # so each plot can be pretty.
        # Note: `lab[0, :, 0]` is the lightness.
        y_ = cmap_as_lab[0, :, 0]
        c_ = x

        x_offset = ind * horizontal_spacing

        # To make scatter markers change color along plot: https://stackoverflow.com/q/8202605/.
        ax.scatter(x + x_offset, y_, c=c_, cmap=cmap, s=markersize, linewidths=0.0)

        # Store locations for colormap labels.
        if cmap_type == "linear":
            xaxis_tick_locations.append(x[-1] + x_offset)
        else:
            xaxis_tick_locations.append(x[int(np.round(steps / 2))] + x_offset)

    # Lightness goes from 0 to 100.
    ax.set_ylim(0.0, 100.0)

    # Set up labels for colormaps.
    ax.xaxis.set_ticks_position("top")
    xaxis_tick_locator = mpl.ticker.FixedLocator(xaxis_tick_locations)  # type: ignore
    ax.xaxis.set_major_locator(xaxis_tick_locator)
    ax.xaxis.set_tick_params(rotation=tickrotation)
    ax.set_xticklabels(labels=xaxis_tick_labels)
    ax.set_ylabel("Lightness $L^*$", fontsize=12)

    if title is not None:
        ax.set_xlabel(title, fontsize=14)

    plt.show()  # type: ignore

    if return_fig:
        return fig


def display_all_gradients() -> None:
    """Displays all registered gradients."""
    for gradient in all_gradients:
        print(gradient.name)
        print(gradient.swatch())


def display_all_palettes() -> None:
    """Displays all registered palettes."""
    for palette in all_palettes:
        print(palette.name)
        print(palette.swatch())



import plotly.express as px
import pytest

import arcadia_pycolor as apc


def simple_plot():
    proteins = ["Protein A", "Protein B", "Protein C"]
    molecular_weights = [50, 75, 100]
    fig = px.bar(
        x=proteins,
        y=molecular_weights,
        labels={"x": "Proteins", "y": "Molecular Weight (kDa)"},
    )
    apc.plotly.style_plot(fig, categorical_axes="x", monospaced_axes="y")
    apc.plotly.set_figure_dimensions(fig, "full_square")
    return fig


@pytest.mark.parametrize(
    "fname, filetypes, expected_outputs",
    [
        ("test.pdf", None, ["test.pdf"]),
        ("test.pdf", ["pdf"], ["test.pdf"]),
        ("test.pdf", ["pdf", "png"], ["test.pdf", "test.png"]),
        ("test.pdf", ["png", "svg"], ["test.pdf", "test.png", "test.svg"]),
        ("test.pdf", ["png"], ["test.pdf", "test.png"]),
        ("test", ["pdf"], ["test.pdf"]),
        ("test", ["svg"], ["test.svg"]),
        ("test", ["pdf", "png"], ["test.pdf", "test.png"]),
    ],
)
def test_plotly_save_figure_filetype_examples(tmp_path, fname, filetypes, expected_outputs):
    """Test the `plotly.save_figure` function with various suffixes."""
    fig = simple_plot()
    apc.plotly.save_figure(
        fig,
        tmp_path / fname,
        "full_square",
        filetypes=filetypes,
    )
    for output in expected_outputs:
        output_path = tmp_path / output
        assert output_path.is_file()


@pytest.mark.parametrize(
    "fname, filetypes",
    [
        ("test.pdf", ["invalid"]),
        ("test", ["invalid"]),
        ("test.invalid", ["pdf"]),
        ("test.invalid", None),
    ],
)
def test_plotly_save_figure_filetype_invalid(tmp_path, fname, filetypes, capsys):
    fig = simple_plot()
    apc.plotly.save_figure(
        fig,
        tmp_path / fname,
        "float_square",
        filetypes=filetypes,
    )
    captured = capsys.readouterr()
    assert "Invalid filetype 'invalid'. Skipping." in captured.out


def test_plotly_save_figure_no_filetype(tmp_path):
    fig = simple_plot()
    with pytest.raises(ValueError):
        apc.plotly.save_figure(
            fig,
            tmp_path / "test",
            "full_wide",
            filetypes=None,
        )



import arcadia_pycolor as apc


def test_mpl_setup():
    """Test that `mpl.setup` can be called multiple times without raising an error.

    TODO: consider adding more complex tests.
    """
    apc.mpl.setup()
    apc.mpl.setup()



import pandas as pd
import pytest
import seaborn as sns
from matplotlib import pyplot as plt
from sklearn import datasets

import arcadia_pycolor as apc


def _load_iris_data():
    iris = datasets.load_iris()
    iris_data = pd.DataFrame(iris.data)  # type: ignore
    iris_data.columns = [name.replace(" (cm)", "") for name in iris.feature_names]  # type: ignore
    iris_data["species"] = iris.target_names[iris.target]  # type: ignore
    return iris_data


DATASET = _load_iris_data()


def plot_seaborn_scatterplot(ax):
    """
    Plot a scatterplot using Seaborn with custom colors from the Arcadia palette.
    """
    colors = {
        "setosa": "apc:aegean",
        "versicolor": apc.amber,
        "virginica": apc.seaweed,
    }
    sns.scatterplot(
        data=DATASET,
        x="sepal length",
        y="sepal width",
        hue="species",
        palette=colors,
        ax=ax,
        s=80,
    )
    apc.mpl.style_plot(monospaced_axes="all")


def plot_seaborn_violinplot(ax):
    """
    Plot a violin plot using Seaborn with custom colors from the Arcadia palette.
    """
    colors = {
        "setosa": "apc:aegean",
        "versicolor": apc.amber,
        "virginica": apc.seaweed,
    }
    sns.violinplot(
        data=DATASET,
        x="species",
        y="sepal width",
        hue="species",
        palette=colors,
        ax=ax,
    )
    apc.mpl.style_plot(categorical_axes="x", monospaced_axes="y")


@pytest.mark.parametrize("figure_size", apc.style_defaults.FIGURE_SIZES_IN_INCHES.keys())
@pytest.mark.parametrize(
    "plotting_function",
    [plot_seaborn_scatterplot, plot_seaborn_violinplot],
)
def test_seaborn_plots(output_dirpath, plotting_function, figure_size):
    fig, ax = plt.subplots(figsize=apc.mpl.get_figure_dimensions(figure_size), layout="constrained")
    plotting_function(ax)
    apc.mpl.save_figure(
        output_dirpath / f"{plotting_function.__name__}_{figure_size}.pdf",
        size=figure_size,
        filetypes=["pdf"],
    )
    plt.close(fig)



import pytest

from arcadia_pycolor import HexCode
from arcadia_pycolor.hexcode import _is_hex_code

VALID_HEXCODES = [
    "#FFFFFF",
    "#FFFFFF",
    "#FFF",
    "#fff",
    "#001",
    "#000",
    "#000000",
    "#000001",
    "#FfFfFf",
    "#f90",
    "#ff9900",
    "#f0a2c3",
]

INVALID_HEXCODES = [
    "#ZZZZZZ",
    "#ZZZ",
    "#FF",
    "apples",
    "123456",
    "ffffff",
    "white",
    "",
    "#",
    "0",
    "F",
    0,
    0.5,
    None,
    "#0000",
]


@pytest.mark.parametrize(
    "hex_string, is_hex_code",
    [(hc, True) for hc in VALID_HEXCODES] + [(hc, False) for hc in INVALID_HEXCODES],
)
def test_is_hex_code(hex_string, is_hex_code):
    assert _is_hex_code(hex_string) == is_hex_code


@pytest.mark.parametrize(
    "hex_code",
    VALID_HEXCODES,
)
def test_valid_hexcode(hex_code):
    assert HexCode("a_hexcode", hex_code) == hex_code


@pytest.mark.parametrize(
    "hex_code",
    INVALID_HEXCODES,
)
def test_invalid_hexcode(hex_code):
    with pytest.raises((ValueError, TypeError)):
        HexCode("a_hexcode", hex_code)


@pytest.mark.parametrize(
    "name, hex_code, swatch",
    [
        (
            "white",
            "#FFFFFF",
            "\x1b[48;2;255;255;255m  \x1b[0m\x1b[38;2;255;255;255m white #FFFFFF\x1b[0m",
        ),
        (
            "aegean",
            "#5088C5",
            "\x1b[48;2;80;136;197m  \x1b[0m\x1b[38;2;80;136;197m aegean #5088C5\x1b[0m",
        ),
    ],
)
def test_hexcode_swatch(name, hex_code, swatch):
    assert HexCode(name, hex_code).swatch() == swatch


@pytest.mark.parametrize(
    "name, hex_code, swatch",
    [
        (
            "white",
            "#FFFFFF",
            "\x1b[48;2;255;255;255m  \x1b[0m\x1b[38;2;255;255;255m white #FFFFFF\x1b[0m",
        ),
        (
            "aegean",
            "#5088C5",
            "\x1b[48;2;80;136;197m  \x1b[0m\x1b[38;2;80;136;197m aegean #5088C5\x1b[0m",
        ),
    ],
)
def test_hexcode_repr(name, hex_code, swatch):
    assert HexCode(name, hex_code).__repr__() == swatch


@pytest.mark.parametrize(
    "name, hex_code, width, min_name_width, swatch",
    [
        (
            "white",
            "#FFFFFF",
            2,
            None,
            "\x1b[48;2;255;255;255m  \x1b[0m\x1b[38;2;255;255;255m white #FFFFFF\x1b[0m",
        ),
        (
            "aegean",
            "#5088C5",
            3,
            10,
            "\x1b[48;2;80;136;197m   \x1b[0m\x1b[38;2;80;136;197m aegean     #5088C5\x1b[0m",
        ),
    ],
)
def test_hexcode_swatch_widths(name, hex_code, width, min_name_width, swatch):
    hexcode = HexCode(name, hex_code)
    assert hexcode.swatch(width, min_name_width) == swatch


@pytest.mark.parametrize(
    "name, hex_code, rgb",
    [
        ("white", "#FFFFFF", [255, 255, 255]),
        ("near-white", "#FFFFFE", [255, 255, 254]),
        ("near-black", "#000001", [0, 0, 1]),
        ("darkbrownish", "#001", [0, 0, 17]),
        ("aegean", "#5088C5", [80, 136, 197]),
    ],
)
def test_hexcode_to_rgb(name, hex_code, rgb):
    assert HexCode(name, hex_code).to_rgb() == rgb


@pytest.mark.parametrize(
    "name, hex_code",
    [
        ("white", "#FFFFFF"),
        ("aegean", "#5088C5"),
    ],
)
def test_hexcode_string(name, hex_code):
    assert str(HexCode(name, hex_code)) == hex_code



import plotly.express as px
import pytest
import seaborn as sns
from matplotlib import pyplot as plt

import arcadia_pycolor as apc

BARPLOT_NUM_READS = [
    1_500_000,
    1_200_000,
    1_000_000,
    400_000,
    1_000_000,
    1_700_000,
    1_300_000,
    600_000,
    1_200_000,
    1_500_000,
    1_600_000,
    300_000,
    900_000,
    1_700_000,
    800_000,
]
BARPLOT_SAMPLE_IDS = [
    "AL 4M",
    "EL 2W",
    "EL 4W",
    "EL 4W WGA",
    "EL 12W",
    "OM 2W",
    "OM 4W",
    "OM 4W WGA",
    "OM 8W",
    "WH 1M",
    "WH 2M",
    "WH 2M WGA",
    "WH 4M",
    "WH 2M Hous",
    "WI 3W",
]
BARPLOT_ERRORS = [
    100000,
    80000,
    70000,
    30000,
    70000,
    120000,
    90000,
    40000,
    80000,
    100000,
    110000,
    20000,
    60000,
    120000,
    50000,
]


def _word_wrap_sample_ids(sample_ids):
    return [sample_id.replace(" ", "\n") for sample_id in sample_ids]


def plot_vertical_barplot_with_matplotlib_with_error_bars(ax):
    """
    Plot a vertical barplot using Matplotlib with error bars.
    """
    sample_ids = _word_wrap_sample_ids(BARPLOT_SAMPLE_IDS)
    plt.bar(sample_ids, BARPLOT_NUM_READS, color=apc.aster)
    apc.mpl.style_plot(ax, monospaced_axes="y")
    apc.mpl.set_xaxis_categorical()
    apc.mpl.add_commas_to_axis_tick_labels(ax.get_yaxis())

    plt.errorbar(sample_ids, BARPLOT_NUM_READS, yerr=BARPLOT_ERRORS, fmt="none", color=apc.crow)
    plt.ylabel("Number of reads")
    plt.xlabel("Sample")


def plot_horizontal_barplot_with_matplotlib_with_error_bars(ax):
    """
    Plot a horizontal barplot using Matplotlib.
    This plot is identical to the vertical version, but with the x and y axes swapped.
    """
    plt.barh(BARPLOT_SAMPLE_IDS, BARPLOT_NUM_READS, color=apc.aster)
    apc.mpl.style_plot(ax, monospaced_axes="x")
    apc.mpl.set_yaxis_categorical()
    apc.mpl.add_commas_to_axis_tick_labels(ax.get_xaxis())
    plt.xticks(rotation=30, ha="right")

    plt.errorbar(
        BARPLOT_NUM_READS, BARPLOT_SAMPLE_IDS, xerr=BARPLOT_ERRORS, fmt="none", color=apc.crow
    )
    plt.xlabel("Number of reads")
    plt.ylabel("Sample")


def plot_vertical_barplot_with_matplotlib_with_categories(ax):
    """
    Plot a vertical barplot using Matplotlib with bar colors based on categories
    derived from the sample IDs.
    """
    sample_categories = [sample_id.split(" ")[0] for sample_id in BARPLOT_SAMPLE_IDS]
    category_to_color = dict(
        zip(set(sample_categories), [apc.aster, apc.aegean, apc.amber, apc.seaweed, apc.rose])
    )
    colors = [category_to_color[category] for category in sample_categories]

    plt.bar(
        _word_wrap_sample_ids(BARPLOT_SAMPLE_IDS),
        BARPLOT_NUM_READS,
        color=colors,
    )
    apc.mpl.style_plot(ax, monospaced_axes="y")
    apc.mpl.set_xaxis_categorical()
    apc.mpl.add_commas_to_axis_tick_labels(ax.get_yaxis())
    plt.ylabel("Number of reads")
    plt.xlabel("Sample")


def plot_vertical_barplot_with_seaborn(ax):
    """
    Plot a vertical barplot using Seaborn.
    """
    sns.barplot(
        x=_word_wrap_sample_ids(BARPLOT_SAMPLE_IDS),
        y=BARPLOT_NUM_READS,
        color=apc.aster,
        # Seaborn by default desaturates the colors. This prevents that.
        saturation=1,
        ax=ax,
    )
    apc.mpl.style_plot(ax, monospaced_axes="y")
    apc.mpl.set_xaxis_categorical()
    apc.mpl.add_commas_to_axis_tick_labels(ax.get_yaxis())
    plt.ylabel("Number of reads")
    plt.xlabel("Sample")


def plot_horizontal_barplot_with_seaborn(ax):
    """
    Plot a horizontal barplot using Seaborn.
    This plot is identical to the vertical version, but with the x and y axes swapped.
    """
    sns.barplot(x=BARPLOT_NUM_READS, y=BARPLOT_SAMPLE_IDS, color=apc.aster, saturation=1, ax=ax)
    apc.mpl.style_plot(ax, monospaced_axes="x")
    apc.mpl.set_yaxis_categorical()
    apc.mpl.add_commas_to_axis_tick_labels(ax.get_xaxis())
    plt.xticks(rotation=30, ha="right")
    plt.xlabel("Number of reads")
    plt.ylabel("Sample")


def plot_vertical_barplot_with_seaborn_with_categories(ax):
    """
    Plot a vertical barplot using Seaborn with bar colors based on categories
    derived from the sample IDs.
    """
    sample_categories = [sample_id.split(" ")[0] for sample_id in BARPLOT_SAMPLE_IDS]
    category_to_color = dict(
        zip(set(sample_categories), [apc.aster, apc.aegean, apc.amber, apc.seaweed, apc.rose])
    )

    sns.barplot(
        x=_word_wrap_sample_ids(BARPLOT_SAMPLE_IDS),
        y=BARPLOT_NUM_READS,
        color=apc.aster,
        hue=sample_categories,
        palette=category_to_color,
        saturation=1,
        ax=ax,
    )

    apc.mpl.style_plot(ax, monospaced_axes="y")
    apc.mpl.set_xaxis_categorical()
    apc.mpl.add_commas_to_axis_tick_labels(ax.get_yaxis())

    plt.xlabel("Number of reads")
    plt.ylabel("Sample")

    legend = ax.get_legend()
    legend.set_title("ID")
    legend.set_bbox_to_anchor((1.06, 1))


def plot_vertical_barplot_with_plotly_with_error_bars():
    """Plot a vertical barplot using Plotly with error bars."""
    fig = px.bar(x=BARPLOT_SAMPLE_IDS, y=BARPLOT_NUM_READS, error_y=BARPLOT_ERRORS)
    fig.update_layout(
        xaxis_title="Sample",
        yaxis_title="Number of reads",
        showlegend=False,
    )
    fig.update_traces(
        marker_color=apc.aster,
        error_y_color=apc.crow,
        error_y_thickness=1.5,
        error_y_width=0,
    )
    apc.plotly.style_plot(fig, categorical_axes="x", monospaced_axes="y")
    apc.plotly.add_commas_to_axis_tick_labels(fig)
    apc.plotly.set_figure_dimensions(fig, "full_wide")
    return fig


def plot_horizontal_barplot_with_plotly_with_error_bars():
    """Plot a horizontal barplot using Plotly with error bars."""
    fig = px.bar(x=BARPLOT_NUM_READS, y=BARPLOT_SAMPLE_IDS, error_x=BARPLOT_ERRORS)
    fig.update_layout(
        xaxis_title="Number of reads",
        yaxis_title="Sample",
        showlegend=False,
    )
    fig.update_traces(
        marker_color=apc.aster,
        error_x_color=apc.crow,
        error_x_thickness=1.5,
        error_x_width=0,
    )
    apc.plotly.style_plot(fig, categorical_axes="y", monospaced_axes="x")
    apc.plotly.add_commas_to_axis_tick_labels(fig)
    apc.plotly.set_figure_dimensions(fig, "half_square")
    return fig


def plot_vertical_barplot_with_plotly_with_categories():
    """
    Plot a vertical barplot using Plotly, with bar colors based on categories
    derived from the sample IDs.
    """
    color_labels = [idx.split(" ")[0] for idx in BARPLOT_SAMPLE_IDS]
    color_dict = {
        label: color
        for label, color in zip(
            set(color_labels), [apc.aster, apc.aegean, apc.amber, apc.seaweed, apc.rose]
        )
    }
    fig = px.bar(
        x=BARPLOT_SAMPLE_IDS,
        y=BARPLOT_NUM_READS,
        color=color_labels,
        color_discrete_map=color_dict,
        title=None,
    )
    fig.update_layout(
        xaxis_title="Sample",
        yaxis_title="Number of reads",
        legend=dict(title="ID", yanchor="top", y=1, xanchor="left", x=1.06),
    )
    apc.plotly.style_plot(fig, categorical_axes="x", monospaced_axes="y")
    apc.plotly.add_commas_to_axis_tick_labels(fig)
    apc.plotly.set_figure_dimensions(fig, "full_wide")
    return fig


@pytest.mark.parametrize(
    "plotting_function",
    [
        plot_vertical_barplot_with_matplotlib_with_error_bars,
        plot_horizontal_barplot_with_matplotlib_with_error_bars,
        plot_vertical_barplot_with_matplotlib_with_categories,
        plot_vertical_barplot_with_seaborn,
        plot_horizontal_barplot_with_seaborn,
        plot_vertical_barplot_with_seaborn_with_categories,
    ],
)
@pytest.mark.parametrize("figure_size", apc.style_defaults.FIGURE_SIZES_IN_INCHES.keys())
def test_barplots(output_dirpath, plotting_function, figure_size):
    fig, ax = plt.subplots(figsize=apc.mpl.get_figure_dimensions(figure_size), layout="constrained")
    plotting_function(ax)
    apc.mpl.save_figure(
        output_dirpath / f"{plotting_function.__name__}_{figure_size}.pdf",
        size=figure_size,
        filetypes=["pdf"],
    )
    plt.close(fig)


@pytest.mark.parametrize(
    "plotting_function",
    [
        plot_vertical_barplot_with_plotly_with_error_bars,
        plot_horizontal_barplot_with_plotly_with_error_bars,
        plot_vertical_barplot_with_plotly_with_categories,
    ],
)
def test_barplots_with_plotly(output_dirpath, plotting_function):
    fig = plotting_function()
    fig.write_html(output_dirpath / f"{plotting_function.__name__}.html")



from typing import Any, Union, cast, overload

import matplotlib as mpl
import numpy as np
from colorspacious import cspace_convert  # type: ignore
from numpy.typing import NDArray

from arcadia_pycolor.gradient import Gradient
from arcadia_pycolor.hexcode import HexCode
from arcadia_pycolor.palette import Palette
from arcadia_pycolor.plot import plot_gradient_lightness

CVD_TYPES = {"d": "deuteranomaly", "p": "protanomaly", "t": "tritanomaly"}


def _create_cvd_dict(cvd_type: str, severity: int = 100) -> dict[str, str]:
    """Creates a dictionary for colorspacious to simulate color vision deficiency.

    Args:
        cvd_type (str): The type of color vision deficiency to simulate.
            Either 'd' for deuteranomaly, 'p' for protanomaly, or 't' for tritanomaly.
        severity (int): The severity of the color vision deficiency, from 0 to 100.
    """
    if cvd_type not in CVD_TYPES:
        raise ValueError(
            "Choose 'd' for deuteranomaly, 'p' for protanomaly, and 't' for tritanomaly."
        )

    clipped_severity = np.clip(severity, 0, 100)
    cvd_space = {"name": "sRGB1+CVD", "cvd_type": CVD_TYPES[cvd_type], "severity": clipped_severity}
    return cvd_space


@overload
def simulate_color(colors: HexCode, cvd_type: str = "d", severity: int = 100) -> HexCode: ...


@overload
def simulate_color(
    colors: list[HexCode], cvd_type: str = "d", severity: int = 100
) -> list[HexCode]: ...


def simulate_color(
    colors: Union[HexCode, list[HexCode]], cvd_type: str = "d", severity: int = 100
) -> Union[HexCode, list[HexCode]]:
    """Simulates color vision deficiency for a single color or list of colors.

    Args:
        colors (HexCode or list[HexCode]): The colors to simulate color vision deficiency on.
        cvd_type (str): The type of color vision deficiency to simulate.
            Either 'd' for deuteranomaly, 'p' for protanomaly, or 't' for tritanomaly.
        severity (int): The severity of the color vision deficiency, from 0 to 100.
    """
    cvd_space = _create_cvd_dict(cvd_type=cvd_type, severity=severity)

    if not isinstance(colors, list):
        colors = [colors]

    cvd_colors: list[HexCode] = []

    for color in colors:
        rgb_color = color.to_rgb()
        cvd_color_name = f"{color.name}_{cvd_type}"
        cvd_color = cast(NDArray[np.int64], cspace_convert(rgb_color, cvd_space, "sRGB1"))
        cvd_color = np.clip(cvd_color / 255, 0, 1)
        hex_code = mpl.colors.to_hex(cvd_color)  # type: ignore
        cvd_hexcode = HexCode(name=cvd_color_name, hex_code=hex_code)
        cvd_colors.append(cvd_hexcode)

    if len(cvd_colors) == 1:
        return cvd_colors[0]
    return cvd_colors


def display_all_color(color: HexCode, severity: int = 100) -> None:
    """Displays all color vision deficiency types for a single color.

    Args:
        color (HexCode): The color to simulate color vision deficiency on.
        severity (int): The severity of the color vision deficiency, from 0 to 100.
    """
    cvd_colors = [color] + [simulate_color(color, cvd_type, severity) for cvd_type in CVD_TYPES]
    for cvd_color in cvd_colors:
        print(cvd_color.swatch())


def simulate_palette(palette: Palette, cvd_type: str = "d", severity: int = 100) -> Palette:
    """Simulates color vision deficiency on a palette.

    Args:
        palette (Palette): The palette on which to simulate color vision deficiency.
        cvd_type (str): The type of color vision deficiency to simulate.
            Either 'd' for deuteranomaly, 'p' for protanomaly, or 't' for tritanomaly.
        severity (int): The severity of the color vision deficiency, from 0 to 100.

    Returns:
        Palette: A new palette with the simulated color vision deficiency.
    """
    cvd_hex_colors = simulate_color(palette.colors, cvd_type=cvd_type, severity=severity)
    cvd_palette = Palette(f"{palette.name}_{cvd_type}", cvd_hex_colors)
    return cvd_palette


def display_all_palette(palette: Palette, severity: int = 100) -> None:
    """
    Displays all color vision deficiency types for a palette.

    Args:
        palette (Palette): The palette on which to simulate color vision deficiency.
        severity (int): The severity of the color vision deficiency, from 0 to 100.
    """
    cvd_palettes = [palette] + [
        simulate_palette(palette, cvd_type, severity) for cvd_type in CVD_TYPES
    ]
    for palette in cvd_palettes:
        print(palette.name)
        print(palette.swatch())


def simulate_gradient(gradient: Gradient, cvd_type: str = "d", severity: int = 100) -> Gradient:
    """Simulates color vision deficiency on a gradient.

    Args:
        gradient (Gradient): The gradient to simulate color vision deficiency on.
        cvd_type (str): The type of color vision deficiency to simulate.
            Either 'd' for deuteranomaly, 'p' for protanomaly, or 't' for tritanomaly.
        severity (int): The severity of the color vision deficiency, from 0 to 100.

    Returns:
        Gradient: A new gradient with the simulated color vision deficiency.
    """
    cvd_hex_colors = simulate_color(gradient.anchor_colors, cvd_type=cvd_type, severity=severity)
    cvd_gradient = Gradient(f"{gradient.name}_{cvd_type}", cvd_hex_colors, gradient.anchor_values)
    return cvd_gradient


def display_all_gradient(gradient: Gradient, severity: int = 100) -> None:
    """Displays all color vision deficiency types for a gradient.

    Args:
        gradient (Gradient): The gradient to display.
        severity (int): The severity of the color vision deficiency, from 0 to 100.
    """
    cvd_gradients = [gradient] + [
        simulate_gradient(gradient, cvd_type, severity) for cvd_type in CVD_TYPES
    ]
    for grad in cvd_gradients:
        print(grad.name)
        print(grad.swatch())


def display_all_gradient_lightness(gradient: Gradient, severity: int = 100, **kwargs: Any) -> None:
    """Displays the lightness of all color vision deficiency types for a gradient.

    Args:
        gradient (Gradient): The gradient to display.
        severity (int): The severity of the color vision deficiency, from 0 to 100.
        **kwargs: Additional keyword arguments to pass to `plot_gradient_lightness`.
    """
    plot_gradient_lightness(
        [gradient] + [simulate_gradient(gradient, cvd_type, severity) for cvd_type in CVD_TYPES],
        **kwargs,
    )



import numpy as np
import pytest

from arcadia_pycolor import Gradient, HexCode
from arcadia_pycolor.colors import black, white

from .test_hexcode import INVALID_HEXCODES


@pytest.mark.parametrize(
    "values",
    [
        [0, 1],
        None,
    ],
)
def test_gradient_from_hexcode_list(values):
    colors = [HexCode("white", "#FFFFFF"), HexCode("black", "#000000")]
    assert Gradient("some-gradient", colors, values).anchor_colors == colors


@pytest.mark.parametrize(
    "values",
    [
        [0, 2],
        [-1, 0],
        [0, 1.1],
        [0.25, 0.5],
        [],
        [0],
        [1],
        [np.nan],
        [None],
        ["0", "1"],
    ],
)
def test_gradient_from_hexcode_list_invalid_values(values):
    with pytest.raises(ValueError):
        Gradient("gradient", [HexCode("white", "#FFFFFF"), HexCode("black", "#000000")], values)


@pytest.mark.parametrize(
    "colors, values",
    [
        (
            [HexCode("white", "#FFFFFF"), HexCode("black", "#000000")],
            [0, 0.5, 1],
        ),
        (
            [HexCode("white", "#FFFFFF"), HexCode("gray", "#CCCCCC"), HexCode("black", "#000000")],
            [0, 1],
        ),
    ],
)
def test_gradient_from_hexcode_list_with_wrong_number_of_values(colors, values):
    with pytest.raises(ValueError):
        Gradient("some_gradient", colors, values)


@pytest.mark.parametrize(
    "invalid_color",
    INVALID_HEXCODES,
)
def test_gradient_from_hexcode_list_invalid_input(invalid_color):
    with pytest.raises(ValueError):
        Gradient("some-gradient", [HexCode("black", "#000000"), invalid_color], [0, 1])


@pytest.mark.parametrize(
    "values",
    [
        [0, 1],
        None,
    ],
)
def test_gradient_from_dict(values):
    assert Gradient.from_dict(
        "some_gradient",
        {"white": "#FFFFFF", "black": "#000000"},
        values,
    ).anchor_colors == [HexCode("white", "#FFFFFF"), HexCode("black", "#000000")]


def test_gradient_swatch():
    expected_swatch = "\x1b[48;2;255;255;255m \x1b[0m\x1b[48;2;243;243;243m \x1b[0m\x1b[48;2;231;231;231m \x1b[0m\x1b[48;2;219;219;219m \x1b[0m\x1b[48;2;207;207;207m \x1b[0m\x1b[48;2;195;195;195m \x1b[0m\x1b[48;2;182;182;182m \x1b[0m\x1b[48;2;170;170;170m \x1b[0m\x1b[48;2;158;158;158m \x1b[0m\x1b[48;2;146;146;146m \x1b[0m\x1b[48;2;134;134;134m \x1b[0m\x1b[48;2;121;121;121m \x1b[0m\x1b[48;2;109;109;109m \x1b[0m\x1b[48;2;97;97;97m \x1b[0m\x1b[48;2;85;85;85m \x1b[0m\x1b[48;2;73;73;73m \x1b[0m\x1b[48;2;60;60;60m \x1b[0m\x1b[48;2;48;48;48m \x1b[0m\x1b[48;2;36;36;36m \x1b[0m\x1b[48;2;24;24;24m \x1b[0m\x1b[48;2;12;12;12m \x1b[0m"  # noqa: E501

    assert (
        Gradient(
            "some_gradient",
            [HexCode("white", "#FFFFFF"), HexCode("black", "#000000")],
            [0, 1],
        ).swatch()
        == expected_swatch
    )


def test_gradient_swatch_steps():
    expected_swatch = "\x1b[48;2;255;255;255m \x1b[0m\x1b[48;2;204;204;204m \x1b[0m\x1b[48;2;153;153;153m \x1b[0m\x1b[48;2;102;102;102m \x1b[0m\x1b[48;2;51;51;51m \x1b[0m"  # noqa: E501
    assert (
        Gradient(
            "some_gradient",
            [HexCode("white", "#FFFFFF"), HexCode("black", "#000000")],
        ).swatch(5)
        == expected_swatch
    )


@pytest.mark.parametrize(
    "name, colors, swatch",
    [
        (
            "my_gradient",
            [HexCode("white", "#FFFFFF"), HexCode("black", "#000000")],
            "\x1b[48;2;255;255;255m \x1b[0m\x1b[48;2;243;243;243m \x1b[0m\x1b[48;2;231;231;231m \x1b[0m\x1b[48;2;219;219;219m \x1b[0m\x1b[48;2;207;207;207m \x1b[0m\x1b[48;2;195;195;195m \x1b[0m\x1b[48;2;182;182;182m \x1b[0m\x1b[48;2;170;170;170m \x1b[0m\x1b[48;2;158;158;158m \x1b[0m\x1b[48;2;146;146;146m \x1b[0m\x1b[48;2;134;134;134m \x1b[0m\x1b[48;2;121;121;121m \x1b[0m\x1b[48;2;109;109;109m \x1b[0m\x1b[48;2;97;97;97m \x1b[0m\x1b[48;2;85;85;85m \x1b[0m\x1b[48;2;73;73;73m \x1b[0m\x1b[48;2;60;60;60m \x1b[0m\x1b[48;2;48;48;48m \x1b[0m\x1b[48;2;36;36;36m \x1b[0m\x1b[48;2;24;24;24m \x1b[0m\x1b[48;2;12;12;12m \x1b[0m\n\x1b[48;2;255;255;255m  \x1b[0m\x1b[38;2;255;255;255m white #FFFFFF\x1b[0m 0.0\n\x1b[48;2;0;0;0m  \x1b[0m\x1b[38;2;0;0;0m black #000000\x1b[0m 1.0",  # noqa: E501
        )
    ],
)
def test_gradient_swatch_repr(name, colors, swatch):
    assert Gradient(name, colors).__repr__() == swatch


@pytest.fixture
def black_to_white_gradient() -> Gradient:
    return Gradient("_", [black, white], [0.0, 1.0])


@pytest.mark.parametrize(
    "values, expected_colors",
    [
        ([0, 1], ["#000000", "#ffffff"]),
        ([0, 0.5, 1], ["#000000", "#808080", "#ffffff"]),
        ([1, 2, 3, 4, 5], ["#000000", "#404040", "#808080", "#c0c0c0", "#ffffff"]),
        ([-1, 0, 1], ["#000000", "#808080", "#ffffff"]),
        ([-3, -2, -1], ["#000000", "#808080", "#ffffff"]),
        ([], []),
    ],
)
def test_map_values_basic_cases(
    black_to_white_gradient: Gradient,
    values: list[float],
    expected_colors: list[str],
):
    assert black_to_white_gradient.map_values(values) == expected_colors


@pytest.mark.parametrize(
    "values, min_value, max_value, expected_colors",
    [
        ([0, 0.5, 1], 0, 1, ["#000000", "#808080", "#ffffff"]),
        ([0, 0.5, 1], 0.25, 0.75, ["#000000", "#808080", "#ffffff"]),
        ([-1, 0.5, 2], 0, 1, ["#000000", "#808080", "#ffffff"]),
        ([0, 10], 0, 20, ["#000000", "#808080"]),
    ],
)
def test_map_values_custom_ranges(
    black_to_white_gradient: Gradient,
    values: list[float],
    min_value: float,
    max_value: float,
    expected_colors: list[str],
):
    assert black_to_white_gradient.map_values(values, min_value, max_value) == expected_colors


def test_map_values_invalid_cases(black_to_white_gradient: Gradient):
    # You can't pass min larger than max
    with pytest.raises(ValueError, match="must be greater than"):
        black_to_white_gradient.map_values([0, 1], min_value=1, max_value=0)

    # Or min equal to max
    with pytest.raises(ValueError, match="must be greater than"):
        black_to_white_gradient.map_values([0, 1], min_value=1, max_value=1)


def test_gradient_num_anchors():
    """Test that a Gradient's num_anchors returns the expected count."""
    colors = [HexCode("white", "#FFFFFF"), HexCode("black", "#000000")]
    gradient = Gradient("test_gradient", colors)
    assert gradient.num_anchors == 2
    assert len(gradient.anchor_colors) == 2
    assert len(gradient.anchor_values) == 2


def test_gradient_anchor_properties():
    """Test that a Gradient's anchors have the expected properties."""
    colors = [HexCode("white", "#FFFFFF"), HexCode("black", "#000000")]
    values = [0.0, 1.0]
    gradient = Gradient("test_gradient", colors, values)

    # Check anchor properties.
    assert gradient.anchors[0].color == colors[0]
    assert gradient.anchors[0].value == values[0]
    assert gradient.anchors[1].color == colors[1]
    assert gradient.anchors[1].value == values[1]

    # Check parent properties.
    assert gradient.anchor_colors == colors
    assert gradient.anchor_values == values



import numpy as np
import pytest
import seaborn as sns
from matplotlib import pyplot as plt

import arcadia_pycolor as apc


@pytest.mark.parametrize("figure_size", apc.style_defaults.FIGURE_SIZES_IN_INCHES.keys())
def test_plot_stacked_barplot(output_dirpath, figure_size):
    """
    Plot a stacked barplot using custom mocked data and colors from the Arcadia palette.
    """
    fig, ax = plt.subplots(figsize=apc.mpl.get_figure_dimensions(figure_size), layout="constrained")
    categories = ["Water", "Salt", "Drug A", "Drug B", "Drug C"]
    subcategories = ["Dead", "Malformed", "Alive"]
    data = np.array(
        [
            [5, 10, 15],
            [7, 14, 21],
            [10, 20, 30],
            [3, 6, 20],
            [23, 4, 7],
        ]
    )

    # Initialize the bottom array to 0.
    bottom = np.zeros(len(categories))

    colors = [apc.denim, apc.canary, apc.seaweed]
    for ind in range(len(subcategories)):
        ax.bar(categories, data[:, ind], bottom=bottom, label=subcategories[ind], color=colors[ind])
        bottom += data[:, ind]

    ax.set_xlabel("Condition")
    ax.set_ylabel("Number of embryos")

    handles, labels = ax.get_legend_handles_labels()
    ax.legend(reversed(handles), reversed(labels))
    legend = ax.get_legend()
    assert legend is not None
    legend.set_title("Viability")

    apc.mpl.style_plot(ax, categorical_axes="x", monospaced_axes="y")
    apc.mpl.save_figure(
        output_dirpath / f"test_plot_stacked_barplot_{figure_size}.pdf",
        size=figure_size,
        filetypes=["pdf"],
    )
    plt.close(fig)


@pytest.mark.parametrize("figure_size", apc.style_defaults.FIGURE_SIZES_IN_INCHES.keys())
def test_plot_multiple_line_plot(output_dirpath, figure_size):
    """
    Plot multiple line plots using custom mocked data and colors from the Arcadia palette.
    Note: the line plots are not stacked; they are independently plotted on separate axes.
    """
    lines = np.random.rand(10, 20)
    fig, axes = plt.subplots(
        nrows=len(lines),
        figsize=apc.mpl.get_figure_dimensions(figure_size),
    )

    colors = [apc.aegean, apc.gray, apc.dragon]
    cmap = apc.Gradient(name="", colors=colors).to_mpl_cmap()

    for ind, line in enumerate(lines):
        ax = axes[ind]  # type: ignore
        color = cmap(ind / len(lines))
        ax.plot(line, color=color)
        apc.mpl.style_plot(axes=ax, monospaced_axes="all")
        ax.set_yticks([])

        if ind != len(lines) - 1:
            ax.spines["bottom"].set_visible(False)
            ax.set_xticks([])
        else:
            ax.set_xlabel("Time (s)")

    fig.supylabel("Brightness")
    fig.subplots_adjust(hspace=0)
    apc.mpl.save_figure(
        output_dirpath / f"test_plot_multiple_line_plot_{figure_size}.pdf",
        size=figure_size,
        filetypes=["pdf"],
    )
    plt.close(fig)


@pytest.mark.parametrize("figure_size", apc.style_defaults.FIGURE_SIZES_IN_INCHES.keys())
def test_plot_heatmaps_with_seaborn(output_dirpath, figure_size):
    """
    Plot a heatmap using Seaborn with a colormap from an Arcadia gradient.
    """
    random_data = np.random.rand(8, 8)

    fig, axes = plt.subplots(
        1, 2, figsize=apc.mpl.get_figure_dimensions(figure_size), layout="constrained"
    )
    sns.heatmap(
        random_data,
        ax=axes[0],  # type: ignore
        square=True,
        cmap="apc:magma",
        cbar_kws={"label": "Intensity"},
        annot=True,
        annot_kws={"fontsize": 12},
    )
    sns.heatmap(
        random_data - 0.5,
        ax=axes[1],  # type: ignore
        square=True,
        cmap="apc:purple_green",
        cbar_kws={"label": "Intensity"},
    )

    for ax in axes:  # type: ignore
        apc.mpl.style_plot(
            ax,
            categorical_axes="all",
            monospaced_axes="all",
            colorbar_exists=True,
        )
        ax.set_xlabel("Category 1")
        ax.set_ylabel("Category 2")

    apc.mpl.save_figure(
        output_dirpath / f"test_plot_heatmaps_with_seaborn_{figure_size}.pdf",
        size=figure_size,
        filetypes=["pdf"],
    )
    plt.close(fig)



from __future__ import annotations
import re
from typing import Any, cast

import matplotlib.colors as mcolors
from colorspacious import cspace_converter  # type: ignore

from arcadia_pycolor.display import colorize


def _is_hex_code(hex_string: Any) -> bool:
    """Checks if a string is a valid HEX code."""
    if not isinstance(hex_string, str):
        return False

    match = re.search(r"^#(?:[0-9a-fA-F]{3}){1,2}$", hex_string)
    if match:
        return True
    return False


class HexCode(str):
    """Represents the HEX code of a color.

    Attributes:
        name (str): The name of the color.
        hex_code (str): The HEX code of the color.
    """

    def __new__(cls, name: str, hex_code: str) -> HexCode:
        """Returns a new HexCode object.

        Args:
            name (str): The name of the color.
            hex_code (str): The HEX code of the color.

        Raises:
            ValueError: If the HEX code is invalid.
        """
        if not _is_hex_code(hex_code):
            raise ValueError(f"Invalid HEX code: {hex_code}")

        obj = str.__new__(cls, hex_code)
        obj.name = name
        obj.hex_code = hex_code
        return obj

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value

    @property
    def hex_code(self) -> str:
        return self._hex_code

    @hex_code.setter
    def hex_code(self, value: str) -> None:
        self._hex_code = value

    def __copy__(self) -> HexCode:
        return HexCode(self.name, self.hex_code)

    def __deepcopy__(self, _: dict) -> HexCode:
        return HexCode(self.name, self.hex_code)

    def to_rgb(self) -> list[int]:
        """Returns a tuple of RGB values for the color."""
        return [int(c * 255) for c in mcolors.to_rgb(self.hex_code)]

    def to_cam02ucs(self) -> list[float]:
        """Returns a tuple of CAM02-UCS values for the color.

        The first value is the lightness (J) and the second and third values are
        the chromaticity coordinates (a: redness-to-greenness, b: blueness-to-yellowness).
        """
        # Convert RGB255 to RGB1.
        rgb = [i / 255 for i in self.to_rgb()]

        # Convert RGB1 to CAM02-UCS.
        cam02ucs = cast(list[float], cspace_converter("sRGB1", "CAM02-UCS")(rgb))

        return cam02ucs

    def swatch(self, width: int = 2, min_name_width: int | None = None) -> str:
        """Returns a color swatch with the specified width and color name width.

        Based on colorir's `swatch` function:
        https://github.com/aleferna12/colorir/blob/2d44e4c/colorir/utils.py#L69.

        Args:
            width (int): The desired width of the color swatch.
            min_name_width (int, optional): The desired width of the color name.
        """
        # Add padding to the color name if necessary.
        # Used when displaying multiple colors in a palette.
        if min_name_width:
            color_name = self.name.ljust(min_name_width)
        else:
            color_name = self.name

        # Creates a block of color with the specified width in monospace characters.
        swatch_text = " " * width
        output = colorize(swatch_text, bg_color=self)

        output += colorize(f" {color_name} {self.hex_code}", fg_color=self)

        return output

    def __repr__(self) -> str:
        return self.swatch()

    def __str__(self) -> str:
        return self.hex_code



import arcadia_pycolor.colors as colors
from arcadia_pycolor.palette import Palette

primary = Palette(
    "Primary",
    [
        colors.aegean,
        colors.amber,
        colors.seaweed,
        colors.canary,
        colors.aster,
        colors.rose,
        colors.vital,
        colors.tangerine,
        colors.lime,
        colors.dragon,
        colors.oat,
        colors.wish,
    ],
)

secondary = Palette(
    "Secondary",
    [
        colors.sky,
        colors.dress,
        colors.taupe,
        colors.denim,
        colors.sage,
        colors.marine,
        colors.mars,
        colors.shell,
    ],
)

neutral = Palette(
    "Neutral",
    [
        colors.gray,
        colors.chateau,
        colors.bark,
        colors.slate,
        colors.charcoal,
        colors.crow,
        colors.forest,
        colors.pitch,
    ],
)

background = Palette(
    "Background",
    [colors.parchment, colors.zephyr, colors.lichen, colors.dawn],
)

primary_ordered = Palette(
    "PrimaryOrdered",
    [
        colors.aegean,
        colors.amber,
        colors.canary,
        colors.lime,
        colors.aster,
        colors.rose,
        colors.seaweed,
        colors.dragon,
        colors.vital,
        colors.tangerine,
        colors.oat,
        colors.wish,
    ],
)

secondary_ordered = Palette(
    "SecondaryOrdered",
    [
        colors.sky,
        colors.dress,
        colors.taupe,
        colors.sage,
        colors.denim,
        colors.mars,
        colors.shell,
        colors.marine,
    ],
)

all_ordered = primary_ordered + secondary_ordered
all_ordered.name = "AllOrdered"

blue_shades = Palette(
    "BlueShades",
    [colors.dusk, colors.lapis, colors.aegean, colors.vital, colors.sky],
)

red_shades = Palette(
    "RedShades",
    [colors.cinnabar, colors.dragon, colors.amber, colors.tangerine, colors.melon],
)

yellow_shades = Palette(
    "YellowShades",
    [colors.umber, colors.mustard, colors.canary, colors.sun, colors.oat],
)

purple_shades = Palette(
    "PurpleShades",
    [colors.concord, colors.tanzanite, colors.aster, colors.wish, colors.iris],
)

teal_shades = Palette(
    "TealShades",
    [colors.depths, colors.asparagus, colors.seaweed, colors.teal, colors.glass],
)

pink_shades = Palette(
    "PinkShades",
    [colors.azalea, colors.candy, colors.rose, colors.dress, colors.putty],
)

warm_gray_shades = Palette(
    "WarmGrayShades",
    [colors.mud, colors.bark, colors.chateau, colors.taupe, colors.stone],
)

cool_gray_shades = Palette(
    "CoolGrayShades",
    [colors.steel, colors.marine, colors.cloud, colors.dove, colors.ice],
)

green_shades = Palette(
    "GreenShades",
    [colors.yucca, colors.fern, colors.matcha, colors.lime, colors.edamame],
)

other = Palette(
    "Other",
    [
        colors.concord,
        colors.brightgray,
        colors.paper,
        colors.redwood,
        colors.depths,
        colors.soil,
        colors.umber,
        colors.parchment,
        colors.heather,
        colors.tumbleweed,
        colors.wheat,
        colors.shire,
        colors.topaz,
        colors.space,
        colors.butter,
        colors.terracotta,
        colors.blush,
        colors.lilac,
        colors.ghost,
    ],
)

named = Palette(
    "Named",
    [
        colors.white,
        colors.red,
        colors.green,
        colors.blue,
        colors.cyan,
        colors.magenta,
        colors.yellow,
        colors.black,
    ],
)

# All palettes
all_colors = primary + secondary + neutral + background + other + named
shades = (
    blue_shades
    + red_shades
    + yellow_shades
    + purple_shades
    + teal_shades
    + pink_shades
    + warm_gray_shades
    + cool_gray_shades
    + green_shades
)
all_colors = all_colors + Palette(
    name="Shades", colors=[color for color in shades.colors if color not in all_colors.colors]
)
all_colors.name = "AllColors"


all_palettes = [obj for obj in globals().values() if isinstance(obj, Palette)]



import matplotlib.pyplot as plt
import pytest

import arcadia_pycolor as apc


def simple_plot():
    plt.figure(figsize=(3, 3))
    plt.plot([1, 2, 3], [1, 2, 3])


@pytest.mark.parametrize(
    "fname, filetypes, expected_outputs",
    [
        ("test.pdf", None, ["test.pdf"]),
        ("test.pdf", ["pdf"], ["test.pdf"]),
        ("test.pdf", ["pdf", "png"], ["test.pdf", "test.png"]),
        ("test.pdf", ["png", "eps"], ["test.pdf", "test.png", "test.eps"]),
        ("test.pdf", ["png"], ["test.pdf", "test.png"]),
        ("test", ["pdf"], ["test.pdf"]),
        ("test", ["pdf", "png"], ["test.pdf", "test.png"]),
    ],
)
def test_mpl_save_figure_filetype_examples(tmp_path, fname, filetypes, expected_outputs):
    """
    Test the `mpl.save_figure` function with various suffixes.
    """

    simple_plot()
    apc.mpl.save_figure(
        tmp_path / fname,
        size="half_square",
        filetypes=filetypes,
    )

    for output in expected_outputs:
        output_path = tmp_path / output
        assert output_path.is_file()


@pytest.mark.parametrize(
    "fname, filetypes",
    [
        ("test.pdf", ["invalid"]),
        ("test", ["invalid"]),
        ("test.invalid", ["pdf"]),
        ("test.invalid", None),
    ],
)
def test_mpl_save_figure_filetype_invalid(tmp_path, fname, filetypes, capsys):
    simple_plot()

    apc.mpl.save_figure(
        tmp_path / fname,
        size="half_square",
        filetypes=filetypes,
    )

    captured = capsys.readouterr()
    assert "Invalid filetype 'invalid'. Skipping." in captured.out


def test_mpl_save_figure_no_filetype(tmp_path):
    simple_plot()

    with pytest.raises(ValueError):
        apc.mpl.save_figure(
            tmp_path / "test",
            size="half_square",
            filetypes=None,
        )



from __future__ import annotations
from dataclasses import dataclass

import matplotlib.colors as mcolors

from arcadia_pycolor.display import colorize
from arcadia_pycolor.hexcode import HexCode
from arcadia_pycolor.palette import Palette
from arcadia_pycolor.utils import (
    NumericSequence,
    distribute_values,
    interpolate_x_values,
    is_monotonic,
    rescale_and_concatenate_values,
)


@dataclass
class Anchor:
    """
    A paired color and position value for a gradient.

    Args:
        color: A HexCode object representing the color
        value: A numeric value between 0 and 1 representing the position in the gradient
    """

    color: HexCode
    value: float


class Gradient:
    """A sequence of colors and their positions that define a continuous color gradient.

    Each color is paired with a numeric value between 0 and 1 that determines its position
    in the gradient. The first color is always at position 0 and the last color at position 1.
    Colors in between are interpolated based on their position values to create a smooth gradient.

    Attributes:
        name (str): The name of the gradient.
        anchors (list[Anchor]): A list of gradient anchors.

    Properties:
        anchor_colors (list[HexCode]):
            The list of HexCodes corresponding to each anchor.
        anchor_values (list[float]):
            The list of values corresponding to each anchor.
    """

    def __init__(self, name: str, colors: list[HexCode], values: list[float] | None = None):
        """Initializes a Gradient.

        Args:
            name: The name of the gradient.
            colors: A list of HexCodes.
            values: An optional list of float values. See class docstring for details.

        Raises:
            ValueError:
                - If there are less than two values.
                - If the values are not integers or floats.
                - If the values are not between 0 and 1.
                - If the first value is not 0 or the last value is not 1.
                - If the number of values is not the same as the number of colors.
        """
        self.name = name

        if not all(isinstance(color, HexCode) for color in colors):
            raise ValueError("All colors must be HexCode objects.")

        if values is not None:
            if len(values) < 2:
                raise ValueError("A gradient must have at least two values.")
            if not all(isinstance(value, (int, float)) for value in values):
                raise ValueError("All values must be integers or floats.")
            if not all(0 <= value <= 1 for value in values):
                raise ValueError("All values must be between 0 and 1.")
            if not values[0] == 0 or not values[-1] == 1:
                raise ValueError("The first value must be 0 and the last value must be 1.")
            if len(colors) != len(values):
                raise ValueError("The number of colors and values must be the same.")
            anchor_values = values
        else:
            anchor_values = distribute_values(len(colors))

        self.anchors = [Anchor(color, value) for color, value in zip(colors, anchor_values)]

    @property
    def anchor_colors(self) -> list[HexCode]:
        return [anchor.color for anchor in self.anchors]

    @property
    def anchor_values(self) -> list[float]:
        return [anchor.value for anchor in self.anchors]

    @property
    def num_anchors(self) -> int:
        """Returns the number of anchors in the gradient"""
        return len(self.anchors)

    @classmethod
    def from_dict(
        cls, name: str, colors: dict[str, str], values: list[float] | None = None
    ) -> Gradient:
        """Creates a gradient from a dictionary of colors and values."""
        hex_codes = [HexCode(name, hex_code) for name, hex_code in colors.items()]
        return cls(name, hex_codes, values)

    def swatch(self, steps: int = 21) -> str:
        """
        Returns a gradient swatch with the specified number of steps.

        Args:
            steps (int): the number of swatches to display in the gradient
        """
        # Calculate the color for each step in the gradient
        cmap = self.to_mpl_cmap()

        # Get the color for each step in the gradient.
        colors = [
            HexCode(name=str(ind), hex_code=mcolors.to_hex(cmap(ind / steps)))
            for ind in range(steps)
        ]

        swatches = [colorize(" ", bg_color=c) for c in colors]

        return "".join(swatches)

    def reverse(self) -> Gradient:
        """Returns a new gradient with the colors and values in reverse order"""
        return Gradient(
            name=f"{self.name}_r",
            colors=self.anchor_colors[::-1],
            values=[1 - value for value in self.anchor_values[::-1]],
        )

    def resample_as_palette(self, steps: int = 5) -> Palette:
        """Returns a resampled gradient as a Palette with the specified number of steps."""
        gradient = self.to_mpl_cmap()
        values = distribute_values(steps)
        colors = [
            HexCode(name=f"{self.name}_{i}", hex_code=mcolors.to_hex(gradient(value)))
            for i, value in enumerate(values)
        ]

        return Palette(
            name=f"{self.name}_resampled_{steps}",
            colors=colors,
        )

    def map_values(
        self,
        values: NumericSequence,
        min_value: float | None = None,
        max_value: float | None = None,
    ) -> list[HexCode]:
        """Maps a sequence of values to their corresponding colors from a gradient.

        Args:
            values (NumericSequence): A sequence of values to map to colors.
            min_value (float, optional):
                Determines which value corresponds to the first color in the spectrum.
                Any values below this minimum are assigned to the first color.
                If not provided, the minimum value of `values` is chosen.
            max_value (float, optional):
                Determines which value corresponds to the last color in the spectrum.
                Any values greater than this maximum are assigned to the last color.
                If not provided, the maximum value of `values` is chosen.

        Returns:
            list[HexCode]: A list of HexCode objects corresponding to the values.
        """
        if not len(values):
            return []

        if min_value is None:
            min_value = min(values)

        if max_value is None:
            max_value = max(values)

        if min_value >= max_value:
            raise ValueError(
                f"max_value ({max_value}) must be greater than min_value ({min_value})."
            )

        cmap = self.to_mpl_cmap()

        normalized_values = [(value - min_value) / (max_value - min_value) for value in values]

        return [HexCode(f"{value}", mcolors.to_hex(cmap(value))) for value in normalized_values]

    def interpolate_lightness(self) -> Gradient:
        """Interpolates the gradient to new values based on lightness."""

        if self.num_anchors < 3:
            raise ValueError("Interpolation requires at least three colors.")
        if not is_monotonic(self.anchor_values):
            raise ValueError("Lightness must be monotonically increasing or decreasing.")

        lightness_values = [color.to_cam02ucs()[0] for color in self.anchor_colors]
        new_values = interpolate_x_values(lightness_values)

        return Gradient(
            name=f"{self.name}_interpolated",
            colors=self.anchor_colors,
            values=new_values,
        )

    def __add__(self, other: Gradient) -> Gradient:
        """Return the sum of two gradients by concatenating their colors and values."""
        # If the first gradient ends with the same color as the start of the second gradient,
        # drop the repeated color.
        offset = int(self.anchor_colors[-1] == other.anchor_colors[0])
        new_colors = self.anchor_colors + other.anchor_colors[offset:]
        new_values = rescale_and_concatenate_values(
            self.anchor_values, other.anchor_values[offset:]
        )

        return Gradient(
            name=f"{self.name}_{other.name}",
            colors=new_colors,
            values=new_values,
        )

    def __repr__(self) -> str:
        longest_name_length = max(len(anchor.color.name) for anchor in self.anchors)

        return "\n".join(
            [self.swatch()]
            + [
                f"{anchor.color.swatch(min_name_width=longest_name_length)} {anchor.value}"
                for anchor in self.anchors
            ]
        )

    def to_mpl_cmap(self) -> mcolors.LinearSegmentedColormap:
        """Converts the gradient to a matplotlib colormap."""
        colors = [(anchor.value, anchor.color.hex_code) for anchor in self.anchors]
        return mcolors.LinearSegmentedColormap.from_list(
            self.name,
            colors=colors,
        )

    def to_plotly_colorscale(self) -> list[tuple[float, str]]:
        """Converts the gradient to a colorscale acceptable by plotly graph objects.

        Example:
        >>> import plotly.graph_objects as go
        >>> import arcadia_pycolor as apc
        >>> gradient = apc.gradients.reds
        >>> data = np.random.rand(10, 10)
        >>> heatmap = go.Heatmap(z=data, colorscale=gradient.to_plotly_colorscale())
        >>> fig = go.Figure(data=[heatmap])
        >>> fig.show()

        Returns:
            list[tuple[tuple, str]]:
                A 256 (8-bit) color scale. Each element is a two-ple of normalized
                position in the colorscale and the associated hex value.
        """
        return [(i / 255.0, mcolors.rgb2hex(self.to_mpl_cmap()(i / 255.0))) for i in range(256)]



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from arcadia_pycolor.hexcode import HexCode


def colorize(
    string: str,
    fg_color: HexCode | None = None,
    bg_color: HexCode | None = None,
) -> str:
    """Colorizes a string with the specified foreground and background colors.

    Based on colorir's `color_str` function. See:
    https://github.com/aleferna12/colorir/blob/2d44e4c/colorir/utils.py#L389.

    Relies on ANSI escape codes for colorization. See:
    https://gist.github.com/fnky/458719343aabd01cfb17a3a4f7296797.

    Args:
        string (str): The string to colorize.
        fg_color (HexCode): The foreground color.
        bg_color (HexCode): The background color.

    Returns:
        str: The colorized string.
    """
    if fg_color:
        rgb = fg_color.to_rgb()
        string = f"\033[38;2;{rgb[0]};{rgb[1]};{rgb[2]}m" + string + "\33[0m"

    if bg_color:
        bg_rgb = bg_color.to_rgb()
        string = f"\033[48;2;{bg_rgb[0]};{bg_rgb[1]};{bg_rgb[2]}m" + string + "\33[0m"

    return string



import arcadia_pycolor


def test_import_arcadia_pycolor():
    assert arcadia_pycolor is not None

    attrs = [
        "colors",
        "gradients",
        "mpl",
        "plotly",
        "palettes",
        "plot",
        "style_defaults",
        "Gradient",
        "HexCode",
        "Palette",
        "aegean",
    ]

    for attr in attrs:
        assert hasattr(arcadia_pycolor, attr)



from textwrap import dedent
from typing import Literal

import matplotlib.pyplot as plt
import plotly.graph_objects as go

import arcadia_pycolor.colors as colors
import arcadia_pycolor.gradients as gradients
import arcadia_pycolor.palettes as palettes

# Units in inches when dpi is 72
BASE_DPI = 72
PRINT_DPI = 300

FIGURE_PADDING_PIXELS = 20
FIGURE_PADDING_INCHES = FIGURE_PADDING_PIXELS / BASE_DPI

# Common figure sizes for Arcadia Creative Cloud library templates.
FigureSize = Literal["full_wide", "full_square", "float_wide", "float_square", "half_square"]

# This dictionary contains figure sizes in inches WITHOUT padding for the transparent border.
# For example, the "full_wide" figure width is 13.33333333 inches, or 960 pixels at 72 DPI.
# They are used to set dimensions for matplotlib figures.
FIGURE_SIZES_IN_INCHES: dict[FigureSize, tuple[float, float]] = {
    "full_wide": (13.33333333, 5.27777777),
    "full_square": (6.52777777, 6.52777777),
    "float_wide": (9.16666667, 5.27777777),
    "float_square": (4.44444444, 4.44444444),
    "half_square": (6.38888888, 6.38888888),
}

# This dictionary contains full figure sizes in pixels WITH padding for the transparent border.
# For example, the "full_wide" figure width is 960 pixels plus 20 pixels of padding on each side.
# They are used to set dimensions for Plotly figures.
FIGURE_SIZES_IN_PIXELS: dict[FigureSize, tuple[int, int]] = {
    "full_wide": (1000, 420),
    "full_square": (500, 500),
    "float_wide": (700, 420),
    "float_square": (360, 360),
    "half_square": (500, 500),
}

# Font families.
FONT_FILTER = "Atkinson"
DEFAULT_FONT = "Atkinson Hyperlegible Next"
MONOSPACE_FONT = "Atkinson Hyperlegible Mono"

DEFAULT_FONT_PLOTLY = "AtkinsonHyperlegibleNext"
MONOSPACE_FONT_PLOTLY = "AtkinsonHyperlegibleMono"

# Font sizes.
BASE_FONT_SIZE = 15
TITLE_FONT_SIZE = 16
MONOSPACE_FONT_SIZE = 14.5

# Specifications for categorical axes.
CATEGORICAL_AXIS_TICKLENGTH = 0
CATEGORICAL_AXIS_TICKPADDING = 10

# Legend.
LEGEND_SEPARATOR_LINEWIDTH = 1.5

# Axes.
NUMERICAL_AXIS_TICKLENGTH = 5
NUMERICAL_AXIS_TICKPADDING = 5
LINEWEIGHT = 0.75

# Matplotlib runtime configuration parameters.
# API reference: https://matplotlib.org/stable/api/matplotlib_configuration_api.html.
ARCADIA_MATPLOTLIB_RC_PARAMS = {
    # Fonts.
    "font.family": "sans-serif",
    "font.size": BASE_FONT_SIZE,
    "font.sans-serif": "Atkinson Hyperlegible Next",
    "font.monospace": "Atkinson Hyperlegible Mono",
    "font.weight": "regular",
    # Figure.
    "figure.titlesize": TITLE_FONT_SIZE,
    "figure.titleweight": "medium",
    "figure.facecolor": colors.white,
    "figure.edgecolor": "none",
    "figure.frameon": False,
    "figure.dpi": BASE_DPI,
    # Axes.
    "axes.facecolor": "none",
    "axes.edgecolor": colors.black,
    "axes.linewidth": LINEWEIGHT,
    "axes.grid": False,
    "axes.grid.axis": "both",
    "axes.grid.which": "major",
    "axes.prop_cycle": plt.cycler(color=palettes.all_ordered.colors),  # type: ignore
    "axes.titlesize": TITLE_FONT_SIZE + 2,
    "axes.titleweight": "medium",
    "axes.titlepad": 16,
    "axes.labelsize": BASE_FONT_SIZE,
    "axes.labelweight": "medium",
    "axes.labelcolor": colors.black,
    "axes.labelpad": 10,
    "axes.spines.left": True,
    "axes.spines.bottom": True,
    "axes.spines.right": False,
    "axes.spines.top": False,
    "axes.xmargin": 0.04,
    "axes.ymargin": 0.04,
    "axes.zmargin": 0.04,
    "axes.autolimit_mode": "data",
    "polaraxes.grid": True,
    "axes3d.grid": True,
    # Ticks.
    "xtick.major.size": NUMERICAL_AXIS_TICKLENGTH,
    "xtick.minor.size": NUMERICAL_AXIS_TICKLENGTH / 2,
    "xtick.major.width": LINEWEIGHT,
    "xtick.minor.width": LINEWEIGHT,
    "xtick.major.pad": NUMERICAL_AXIS_TICKPADDING,
    "xtick.minor.pad": NUMERICAL_AXIS_TICKPADDING,
    "xtick.color": colors.black,
    "xtick.labelsize": BASE_FONT_SIZE,
    "ytick.major.size": NUMERICAL_AXIS_TICKLENGTH,
    "ytick.minor.size": NUMERICAL_AXIS_TICKLENGTH / 2,
    "ytick.major.width": LINEWEIGHT,
    "ytick.minor.width": LINEWEIGHT,
    "ytick.major.pad": NUMERICAL_AXIS_TICKPADDING,
    "ytick.minor.pad": NUMERICAL_AXIS_TICKPADDING,
    "ytick.color": colors.black,
    "ytick.labelsize": BASE_FONT_SIZE,
    # Legend.
    "legend.loc": "best",
    "legend.frameon": False,
    "legend.title_fontsize": TITLE_FONT_SIZE,
    "legend.fontsize": BASE_FONT_SIZE,
    "legend.framealpha": 0,
    "legend.borderpad": 0,
    "legend.borderaxespad": 0,
    "legend.facecolor": "none",
    "legend.edgecolor": "none",
    "legend.handlelength": 1,
    "legend.handleheight": 1.2,
    "legend.handletextpad": 0.4,
    # Lines.
    "lines.linewidth": 2,
    "lines.linestyle": "-",
    "lines.color": colors.aegean,
    "lines.marker": "none",
    "lines.markerfacecolor": "auto",
    "lines.markeredgecolor": "auto",
    "lines.markeredgewidth": 0,
    "lines.markersize": 6,
    "lines.antialiased": True,
    "lines.dash_joinstyle": "round",
    "lines.dash_capstyle": "butt",
    "lines.solid_joinstyle": "round",
    "lines.solid_capstyle": "round",
    # Markers.
    "markers.fillstyle": "full",
    "pcolor.shading": "auto",
    # Patches.
    "patch.linewidth": 0,  # Removes edge on patches.
    "patch.facecolor": colors.aegean,
    "patch.edgecolor": colors.aegean,
    "patch.force_edgecolor": False,
    "patch.antialiased": True,
    # Saving figures.
    "savefig.format": "pdf",
    "savefig.transparent": True,
    "savefig.pad_inches": FIGURE_PADDING_INCHES,
    "savefig.dpi": BASE_DPI,
    "pdf.fonttype": 42,
    "pdf.compression": 0,
    "ps.fonttype": 42,
    "svg.fonttype": "none",
    "svg.image_inline": True,
    # Images.
    "image.cmap": f"apc:{gradients.magma.name}",
    "image.aspect": "equal",
    "image.interpolation": "antialiased",
    "image.lut": 256,
    "image.origin": "upper",
    "image.resample": True,
    "image.composite_image": True,
}

# Plotly template layout.
# API reference: https://plotly.com/python-api-reference/generated/plotly.graph_objects.layout.html.
PLOTLY_3D_AXIS_ATTRIBUTES = dict(
    linecolor="black",
    linewidth=1,
    showgrid=True,
    showline=True,
    ticks="outside",
    tickwidth=1,
    title=dict(
        font=dict(family=f"{DEFAULT_FONT_PLOTLY}-Medium", size=BASE_FONT_SIZE, color="black"),
    ),
)

ARCADIA_PLOTLY_TEMPLATE_LAYOUT = go.Layout(
    bargap=0.20,
    coloraxis=go.layout.Coloraxis(
        colorbar=go.layout.coloraxis.ColorBar(
            outlinecolor="white",
            thickness=15,
            ticks="outside",
            tickfont=dict(family=MONOSPACE_FONT_PLOTLY, size=MONOSPACE_FONT_SIZE),
            title=dict(
                font=dict(family=f"{DEFAULT_FONT_PLOTLY}-Medium", size=BASE_FONT_SIZE),
                side="right",
            ),
        ),
    ),
    colorscale=go.layout.Colorscale(
        sequential=gradients.magma.to_plotly_colorscale(),
        sequentialminus=gradients.magma.reverse().to_plotly_colorscale(),
        diverging=gradients.orange_sage.to_plotly_colorscale(),
    ),
    font=go.layout.Font(family=DEFAULT_FONT_PLOTLY, size=BASE_FONT_SIZE, color="black"),
    hoverlabel=go.layout.Hoverlabel(
        font_family=DEFAULT_FONT_PLOTLY,
        font_size=13,
    ),
    legend=go.layout.Legend(
        title=dict(
            font=dict(family=f"{DEFAULT_FONT_PLOTLY}-SemiBold", size=TITLE_FONT_SIZE, color="black")
        ),
        font=dict(family=DEFAULT_FONT_PLOTLY, size=BASE_FONT_SIZE, color="black"),
        indentation=-12,
        xanchor="right",
        x=1,
        yanchor="top",
        y=1,
    ),
    margin=go.layout.Margin(
        l=FIGURE_PADDING_PIXELS + 75,
        b=FIGURE_PADDING_PIXELS + 65,
        r=FIGURE_PADDING_PIXELS + 20,
        t=FIGURE_PADDING_PIXELS + 20,
    ),
    scene=go.layout.Scene(
        xaxis=go.layout.scene.XAxis(**PLOTLY_3D_AXIS_ATTRIBUTES),
        yaxis=go.layout.scene.YAxis(**PLOTLY_3D_AXIS_ATTRIBUTES),
        zaxis=go.layout.scene.ZAxis(**PLOTLY_3D_AXIS_ATTRIBUTES),
    ),
    title=go.layout.Title(
        font=dict(family=f"{DEFAULT_FONT_PLOTLY}-SemiBold", size=TITLE_FONT_SIZE, color="black"),
        automargin=True,
        yref="container",
    ),
    xaxis=go.layout.XAxis(
        automargin=True,
        linecolor="black",
        linewidth=1,
        showgrid=False,
        showline=True,
        ticklabelstandoff=2,
        ticks="outside",
        tickwidth=1,
        title=dict(
            font=dict(family=f"{DEFAULT_FONT_PLOTLY}-Medium", size=BASE_FONT_SIZE, color="black"),
            standoff=10,
        ),
        zerolinecolor="rgba(0,0,0,0)",
        zerolinewidth=0,
    ),
    yaxis=go.layout.YAxis(
        automargin=True,
        linecolor="black",
        linewidth=1,
        showgrid=False,
        showline=True,
        ticklabelstandoff=2,
        ticks="outside",
        tickwidth=1,
        title=dict(
            font=dict(family=f"{DEFAULT_FONT_PLOTLY}-Medium", size=BASE_FONT_SIZE, color="black"),
            standoff=10,
        ),
        zerolinecolor="rgba(0,0,0,0)",
        zerolinewidth=0,
    ),
)

PLOTLY_HTML_EXPORT_CSS = dedent(
    """
    @font-face {
      font-family: "AtkinsonHyperlegibleNext";
      src: url("https://fonts.gstatic.com/s/atkinsonhyperlegiblenext/v7/NaP4cYPdHfdVxJw0IfIP0lvYFqijb-UxCtm5_wdGscKFt4tOOfV4ZmW37LQhtA.ttf")
        format("truetype");
      font-weight: normal;
      font-style: normal;
    }

    @font-face {
      font-family: "AtkinsonHyperlegibleNext-Medium";
      src: url("https://fonts.gstatic.com/s/atkinsonhyperlegiblenext/v7/NaP4cYPdHfdVxJw0IfIP0lvYFqijb-UxCtm5_wdGscKFt4tOOfV4ZmW33rQhtA.ttf")
        format("truetype");
      font-weight: normal;
      font-style: normal;
    }

    @font-face {
      font-family: "AtkinsonHyperlegibleNext-SemiBold";
      src: url("https://fonts.gstatic.com/s/atkinsonhyperlegiblenext/v7/NaP4cYPdHfdVxJw0IfIP0lvYFqijb-UxCtm5_wdGscKFt4tOOfV4ZmW3MrMhtA.ttf")
        format("truetype");
      font-weight: normal;
      font-style: normal;
    }

    @font-face {
      font-family: "AtkinsonHyperlegibleMono";
      src: url("https://fonts.gstatic.com/s/atkinsonhyperlegiblemono/v8/tssNAoFBci4C4gvhPXrt3wjT1MqSzhA4t7IIcncBiyihrK15gZ4k_SaZHNeiDQ.ttf")
        format("truetype");
      font-weight: normal;
      font-style: normal;
    }
    """
)



from __future__ import annotations
from typing import overload

import matplotlib.colors as mcolors

from arcadia_pycolor.display import colorize
from arcadia_pycolor.hexcode import HexCode


class Palette:
    """A discrete ordered sequence of HexCode objects.

    Attributes:
        name (str): The name of the palette.
        colors (list): A list of HexCode objects.
    """

    def __init__(self, name: str, colors: list[HexCode]):
        """
        Initialize a Palette with a name and a list of colors.

        Args:
            name (str): the name of the palette
            colors (list): a list of HexCode objects
        """
        self.name = name

        if not all(isinstance(color, HexCode) for color in colors):
            raise ValueError("All colors must be HexCode objects.")

        self.colors = colors

    @classmethod
    def from_dict(cls, name: str, colors: dict[str, str]) -> Palette:
        """Create a palette from a dictionary of color names and hex codes.

        Args:
            name (str): The name of the palette.
            colors (dict): A dictionary of color names and hex codes.
        """
        hex_codes = [HexCode(name, hex_code) for name, hex_code in colors.items()]
        return cls(name, hex_codes)

    def swatch(self) -> str:
        """Returns a swatch of the palette."""
        swatches = [colorize("  ", bg_color=color) for color in self.colors]
        return "".join(swatches)

    def reverse(self) -> Palette:
        """Returns a reversed palette."""
        return Palette(
            name=f"{self.name}_r",
            colors=self.colors[::-1],
        )

    def __len__(self) -> int:
        """Returns the number of colors in the palette."""
        return len(self.colors)

    def __iter__(self):
        """Returns an iterator over the colors in the palette."""
        return iter(self.colors)

    @overload
    def __getitem__(self, index: int) -> HexCode: ...

    @overload
    def __getitem__(self, index: slice) -> Palette: ...

    def __getitem__(self, index: int | slice) -> HexCode | Palette:
        """Returns the color at the given index, or a new palette if a slice is provided.

        Args:
            index: An integer index or slice.

        Returns:
            A HexCode if an integer index was provided, or a new Palette if
            a slice was provided.
        """
        if isinstance(index, slice):
            return Palette(name=f"{self.name}_slice", colors=self.colors[index])
        else:
            return self.colors[index]

    def __repr__(self) -> str:
        """Returns a string representation of the palette."""
        longest_name_length = max(len(color.name) for color in self.colors)

        return "\n".join(
            [self.swatch()]
            + [color.swatch(min_name_width=longest_name_length) for color in self.colors]
        )

    def __add__(self, other: Palette) -> Palette:
        """Returns a new palette that is the concatenation of this palette and another."""
        return Palette(
            name=f"{self.name}+{other.name}",
            colors=self.colors + other.colors,
        )

    def to_mpl_cmap(self) -> mcolors.ListedColormap:
        """Returns a matplotlib colormap for the palette."""
        return mcolors.ListedColormap([color.hex_code for color in self.colors], self.name)



from arcadia_pycolor.hexcode import HexCode

# Primary colors.
aegean = HexCode("aegean", "#5088C5")
amber = HexCode("amber", "#F28360")
seaweed = HexCode("seaweed", "#3B9886")
canary = HexCode("canary", "#F7B846")
aster = HexCode("aster", "#7A77AB")
rose = HexCode("rose", "#F898AE")
vital = HexCode("vital", "#73B5E3")
tangerine = HexCode("tangerine", "#FFB984")
lime = HexCode("lime", "#97CD78")
dragon = HexCode("dragon", "#C85152")
oat = HexCode("oat", "#F5E4BE")
wish = HexCode("wish", "#BABEE0")

# Secondary colors.
sky = HexCode("sky", "#C6E7F4")
dress = HexCode("dress", "#F8C5C1")
taupe = HexCode("taupe", "#DBD1C3")
denim = HexCode("denim", "#B6C8D4")
sage = HexCode("sage", "#B5BEA4")
marine = HexCode("marine", "#8A99AD")
mars = HexCode("mars", "#DA9085")
shell = HexCode("shell", "#EDE0D6")

# Neutral colors.
gray = HexCode("gray", "#EBEDE8")
chateau = HexCode("chateau", "#BAB0A8")
bark = HexCode("bark", "#8F8885")
slate = HexCode("slate", "#43413F")
charcoal = HexCode("charcoal", "#484B50")
crow = HexCode("crow", "#292928")
forest = HexCode("forest", "#596F74")
pitch = HexCode("pitch", "#09090A")

# Background colors.
parchment = HexCode("parchment", "#FDF8F2")
zephyr = HexCode("zephyr", "#F4FBFF")
lichen = HexCode("lichen", "#F7FBEF")
dawn = HexCode("dawn", "#F8F4F1")

# Shades.
lapis = HexCode("lapis", "#2B65A1")
dusk = HexCode("dusk", "#094468")
cinnabar = HexCode("cinnabar", "#9E3F41")
melon = HexCode("melon", "#FFD0B0")
sun = HexCode("sun", "#FFD364")
mustard = HexCode("mustard", "#D68D22")
umber = HexCode("umber", "#A85E28")
iris = HexCode("iris", "#DCDFEF")
tanzanite = HexCode("tanzanite", "#54448C")
glass = HexCode("glass", "#C3E2DB")
teal = HexCode("teal", "#6FBCAD")
asparagus = HexCode("asparagus", "#2A6B5E")
depths = HexCode("depths", "#09473E")
putty = HexCode("putty", "#FFE3D4")
candy = HexCode("candy", "#E2718F")
azalea = HexCode("azalea", "#C14C70")
stone = HexCode("stone", "#EDE6DA")
mud = HexCode("mud", "#635C5A")
ice = HexCode("ice", "#E6EAED")
dove = HexCode("dove", "#CAD4DB")
cloud = HexCode("cloud", "#ABBAC4")
steel = HexCode("steel", "#687787")
fern = HexCode("fern", "#47784A")
edamame = HexCode("edamame", "#C1E1AE")
matcha = HexCode("matcha", "#71AC5A")
yucca = HexCode("yucca", "#1E4812")

# Gradient colors.
concord = HexCode("concord", "#341E60")
heather = HexCode("heather", "#A96789")
tumbleweed = HexCode("tumbleweed", "#E9A482")
wheat = HexCode("wheat", "#F5DFB2")
shire = HexCode("shire", "#4E7F72")
topaz = HexCode("topaz", "#FFCC7B")
space = HexCode("space", "#282A49")
butter = HexCode("butter", "#FFFDBD")
terracotta = HexCode("terracotta", "#964222")
blush = HexCode("blush", "#FFF3F4")
lilac = HexCode("lilac", "#6862AB")
ghost = HexCode("ghost", "#FCF7FF")
redwood = HexCode("redwood", "#52180A")

# Other Arcadia colors.
brightgray = HexCode("brightgray", "#EAEAEA")
paper = HexCode("paper", "#FCFCFC")

soil = HexCode("soil", "#4D2500")
blossom = HexCode("blossom", "#F5CBE4")
mint = HexCode("mint", "#D1EADF")

# Other named colors.
white = HexCode("white", "#FFFFFF")
red = HexCode("red", "#FF0000")
green = HexCode("green", "#00FF00")
blue = HexCode("blue", "#0000FF")
cyan = HexCode("cyan", "#00FFFF")
magenta = HexCode("magenta", "#FF00FF")
yellow = HexCode("yellow", "#FFFF00")
black = HexCode("black", "#000000")



from pathlib import Path
from typing import Any, Literal, Union, get_args

import plotly.graph_objects as go
import plotly.io as pio
from bs4 import BeautifulSoup

from arcadia_pycolor.style_defaults import (
    ARCADIA_PLOTLY_TEMPLATE_LAYOUT,
    DEFAULT_FONT_PLOTLY,
    FIGURE_SIZES_IN_PIXELS,
    MONOSPACE_FONT_PLOTLY,
    MONOSPACE_FONT_SIZE,
    PLOTLY_HTML_EXPORT_CSS,
    FigureSize,
)

# Reference: https://plotly.com/python/3d-charts/.
PLOTLY_3D_TRACE_TYPES = (
    go.Mesh3d,
    go.Scatter3d,
    go.Cone,
    go.Isosurface,
    go.Surface,
    go.Volume,
    go.Contour,
    go.Parcoords,
    go.Streamtube,
)

AxisSelector = Literal["x", "y", "z", "xy", "yz", "xz", "xyz", "all"]


def _has_subplots(fig: go.Figure) -> bool:
    layout_keys = fig.layout.to_plotly_json().keys()  # type: ignore
    xaxes = [key for key in layout_keys if key.startswith("xaxis")]
    yaxes = [key for key in layout_keys if key.startswith("yaxis")]
    return len(xaxes) > 1 or len(yaxes) > 1


def _is_3d_plot(fig: go.Figure, row: Union[int, None] = None, col: Union[int, None] = None) -> bool:
    """Returns True if the figure data only contains 3D traces."""
    if _has_subplots(fig) and row is not None and col is not None:
        return isinstance(fig.get_subplot(row, col), go.layout.Scene)
    return isinstance(fig.data[0], PLOTLY_3D_TRACE_TYPES)


def _is_plot_with_3d_traces(fig: go.Figure) -> bool:
    """Returns True if the figure data contains any 3D traces."""
    return any(isinstance(trace, PLOTLY_3D_TRACE_TYPES) for trace in fig.data)


def _is_plot_with_3d_traces_only(fig: go.Figure) -> bool:
    """Returns True if the figure data only contains 3D traces."""
    return all(isinstance(trace, PLOTLY_3D_TRACE_TYPES) for trace in fig.data)


def _is_plot_with_colorbar(fig: go.Figure) -> bool:
    """Returns True if the figure layout contains a non-emptycolorbar."""
    return len(fig.layout.coloraxis.colorbar.to_plotly_json()) > 0  # type: ignore


def _is_plot_with_legend(fig: go.Figure) -> bool:
    """Returns True if the figure layout contains a non-empty legend."""
    return len(fig.layout.legend.to_plotly_json()) > 0  # type: ignore


def _revert_to_default_fonts(fig: go.Figure) -> None:
    """Reverts the fonts in a Plotly figure to the default Plotly fonts."""
    template_without_fonts = go.Layout(**ARCADIA_PLOTLY_TEMPLATE_LAYOUT.to_plotly_json())

    template_without_fonts.update(
        font_family=None,
        title_font_family=None,
        legend_title_font_family=None,
        legend_font_family=None,
        hoverlabel_font_family=None,
        coloraxis_colorbar_title_font_family=None,
        coloraxis_colorbar_tickfont_family="monospace",
        xaxis_title_font_family=None,
        yaxis_title_font_family=None,
        scene_xaxis_title_font_family=None,
        scene_yaxis_title_font_family=None,
        scene_zaxis_title_font_family=None,
    )

    is_monospaced_xaxis = fig.layout.xaxis.tickfont.family == MONOSPACE_FONT_PLOTLY  # type: ignore
    is_monospaced_yaxis = fig.layout.yaxis.tickfont.family == MONOSPACE_FONT_PLOTLY  # type: ignore

    fig.update_layout(
        xaxis_tickfont_family="monospace" if is_monospaced_xaxis else None,
        yaxis_tickfont_family="monospace" if is_monospaced_yaxis else None,
        xaxis_tickfont_size=13.5 if is_monospaced_xaxis else None,
        yaxis_tickfont_size=13.5 if is_monospaced_yaxis else None,
    )

    is_monospaced_xaxis = fig.layout.scene.xaxis.tickfont.family == MONOSPACE_FONT_PLOTLY  # type: ignore
    is_monospaced_yaxis = fig.layout.scene.yaxis.tickfont.family == MONOSPACE_FONT_PLOTLY  # type: ignore
    is_monospaced_zaxis = fig.layout.scene.zaxis.tickfont.family == MONOSPACE_FONT_PLOTLY  # type: ignore

    fig.update_scenes(
        xaxis_title_font_family=None,
        yaxis_title_font_family=None,
        zaxis_title_font_family=None,
        xaxis_tickfont_family="monospace" if is_monospaced_xaxis else None,
        yaxis_tickfont_family="monospace" if is_monospaced_yaxis else None,
        zaxis_tickfont_family="monospace" if is_monospaced_zaxis else None,
        xaxis_tickfont_size=13.5 if is_monospaced_xaxis else None,
        yaxis_tickfont_size=13.5 if is_monospaced_yaxis else None,
        zaxis_tickfont_size=13.5 if is_monospaced_zaxis else None,
    )

    fig.update_layout(template=dict(layout=template_without_fonts))


def _add_fonts_to_plotly_html_export(filepath: str) -> None:
    """Adds a style tag with fonts loaded from Google Fonts to an HTML file's head section.

    This is necessary for embeds of Plotly HTML exports to use the Atkinson Hyperlegible fonts.

    Args:
        filepath (str): Path to the HTML file to modify.
    """
    with open(filepath) as f:
        content = f.read()

    soup = BeautifulSoup(content, "html.parser")

    style_tag = soup.new_tag("style")
    style_tag.string = PLOTLY_HTML_EXPORT_CSS

    if soup.head is None:
        raise ValueError("Could not find <head> tag in HTML file.")
    soup.head.append(style_tag)

    with open(filepath, "w") as f:
        f.write(str(soup))


def save_figure(
    fig: go.Figure,
    filepath: str,
    size: FigureSize,
    filetypes: Union[list[str], None] = None,
    **write_image_kwargs: dict[Any, Any],
) -> None:
    """Saves the current figure to a file without any margins or padding.

    Args:
        fig (go.Figure): The figure to save.
        filepath (str): The path to save the figure to.
        size (FigureSize): The size of the figure, which must be one of the following:
            - "full_wide"
            - "full_square"
            - "float_wide"
            - "float_square"
            - "half_square"
        filetypes (list[str], optional): The file types(s) to save the figure to.
            If None, the original filetype of `filepath` is used.
            If the original filetype is not in `filetypes`, it is appended to the list.
        **write_image_kwargs: Additional keyword arguments to pass to `fig.write_image`.
    """
    # By default, our Plotly template attempts to add 40 pixels of margin on all sides.
    # However, due to Plotly's internal automargin strategy, the margin is not always
    # applied correctly, resulting in a figure that is not the correct size.
    #
    # For exports, we want to remove the margins and update the figure dimensions
    # so that the correct margins can be applied in Adobe Illustrator.
    # TODO(#69): Write a custom function to apply the margins.
    updated_margins = dict(l=0, r=0, t=0, b=0)
    updated_width = FIGURE_SIZES_IN_PIXELS[size][0] - 80
    updated_height = FIGURE_SIZES_IN_PIXELS[size][1] - 80

    # For some reason, the axis linewidths (which are set to 1 px) are being rendered as
    # 1 pt in Illustrator. Manually setting these to 0.75 px renders them as 0.75 pt.
    updated_axis_linewidth = 0.75

    fig_export = go.Figure(fig)
    fig_export.update_layout(
        margin=updated_margins,
        width=updated_width,
        height=updated_height,
        xaxis=dict(linewidth=updated_axis_linewidth),
        yaxis=dict(linewidth=updated_axis_linewidth),
    )

    # If no file types are provided, use the filetype from the file path.
    valid_filetypes = ["png", "jpg", "jpeg", "webp", "svg", "pdf"]

    filename = Path(filepath).with_suffix("")
    filetype = Path(filepath).suffix[1:]

    if filetypes is None:
        if not filetype:
            raise ValueError("The filename must include a filetype if no filetypes are provided.")
        filetypes = [filetype]
    else:
        filetypes.append(filetype)

    for ftype in filetypes:
        if ftype not in valid_filetypes:
            print(f"Invalid filetype '{ftype}'. Skipping.")
            continue
        fig_export.write_image(f"{filename}.{ftype}", **write_image_kwargs)


def export_to_html(fig: go.Figure, filepath: str) -> None:
    """
    Exports the current figure to an HTML file and adds fonts loaded from Google Fonts,
    allowing the figure to be embedded on webpages without requiring fonts to be installed.

    If the figure contains 3D traces, all fonts are reverted to default Plotly fonts.
    This is because HTML exports of 3D plots do not apply remotely loaded fonts correctly.
    See this issue for more details: https://github.com/plotly/plotly.js/issues/7413.

    Args:
        fig (go.Figure): The figure to export.
        filepath (str): The path to save the figure to.
    """

    if _is_plot_with_3d_traces(fig):
        _revert_to_default_fonts(fig)
        fig.write_html(filepath)
    else:
        fig.write_html(filepath)
        _add_fonts_to_plotly_html_export(filepath)


def set_yticklabel_font(
    fig: go.Figure,
    font: str = DEFAULT_FONT_PLOTLY,
    font_size: Union[float, None] = None,
    row: Union[int, None] = None,
    col: Union[int, None] = None,
) -> None:
    """Sets the font and font size of the y-axis tick labels for a Plotly figure.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        font (str): The font family to use for the tick labels.
        font_size (float, optional): The font size to use for the tick labels.
            If None, keeps current size.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    if _is_3d_plot(fig, row, col):
        fig.update_scenes(yaxis_tickfont_family=font, row=row, col=col)
        if font_size is not None:
            fig.update_scenes(yaxis_tickfont_size=font_size, row=row, col=col)
    else:
        fig.update_yaxes(tickfont_family=font, row=row, col=col)
        if font_size is not None:
            fig.update_yaxes(tickfont_size=font_size, row=row, col=col)


def set_xticklabel_font(
    fig: go.Figure,
    font: str = DEFAULT_FONT_PLOTLY,
    font_size: Union[float, None] = None,
    row: Union[int, None] = None,
    col: Union[int, None] = None,
) -> None:
    """Sets the font and font size of the x-axis tick labels for a Plotly figure.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        font (str): The font family to use for the tick labels.
        font_size (float, optional): The font size to use for the tick labels.
            If None, keeps current size.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    if _is_3d_plot(fig, row, col):
        fig.update_scenes(xaxis_tickfont_family=font, row=row, col=col)
        if font_size is not None:
            fig.update_scenes(xaxis_tickfont_size=font_size, row=row, col=col)
    else:
        fig.update_xaxes(tickfont_family=font, row=row, col=col)
        if font_size is not None:
            fig.update_xaxes(tickfont_size=font_size, row=row, col=col)


def set_zticklabel_font(
    fig: go.Figure,
    font: str = DEFAULT_FONT_PLOTLY,
    font_size: Union[float, None] = None,
    row: Union[int, None] = None,
    col: Union[int, None] = None,
) -> None:
    """Sets the font and font size of the z-axis tick labels for a Plotly figure.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        font (str): The font family to use for the tick labels.
        font_size (float, optional): The font size to use for the tick labels.
            If None, keeps current size.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    fig.update_scenes(zaxis_tickfont_family=font, row=row, col=col)
    if font_size is not None:
        fig.update_scenes(zaxis_tickfont_size=font_size, row=row, col=col)


def set_ticklabel_font(
    fig: go.Figure,
    font: str = DEFAULT_FONT_PLOTLY,
    font_size: Union[float, None] = None,
    row: Union[int, None] = None,
    col: Union[int, None] = None,
) -> None:
    """Sets the font and font size of the ticklabels for the given axes.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        font (str, optional): The font family to use for the ticklabels.
        font_size (float, optional): The font size to use for the ticklabels.
            If None, keeps current size.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    set_xticklabel_font(fig, font, font_size, row, col)
    set_yticklabel_font(fig, font, font_size, row, col)
    if _is_3d_plot(fig, row, col):
        set_zticklabel_font(fig, font, font_size, row, col)


def set_xticklabel_monospaced(
    fig: go.Figure, row: Union[int, None] = None, col: Union[int, None] = None
) -> None:
    """Sets the font of the x-axis ticklabels to the default monospace font.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    set_xticklabel_font(fig, MONOSPACE_FONT_PLOTLY, font_size=MONOSPACE_FONT_SIZE, row=row, col=col)


def set_yticklabel_monospaced(
    fig: go.Figure, row: Union[int, None] = None, col: Union[int, None] = None
) -> None:
    """Sets the font of the y-axis ticklabels to the default monospace font.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    set_yticklabel_font(fig, MONOSPACE_FONT_PLOTLY, font_size=MONOSPACE_FONT_SIZE, row=row, col=col)


def set_zticklabel_monospaced(
    fig: go.Figure, row: Union[int, None] = None, col: Union[int, None] = None
) -> None:
    """Sets the font of the z-axis ticklabels to the default monospace font.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    set_zticklabel_font(fig, MONOSPACE_FONT_PLOTLY, font_size=MONOSPACE_FONT_SIZE, row=row, col=col)


def set_colorbar_ticklabel_monospaced(
    fig: go.Figure, row: Union[int, None] = None, col: Union[int, None] = None
) -> None:
    """Sets the font of the colorbar ticklabels to the default monospace font.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    fig.update_coloraxes(
        colorbar_tickfont_family=MONOSPACE_FONT_PLOTLY,
        colorbar_tickfont_size=MONOSPACE_FONT_SIZE,
        row=row,
        col=col,
    )


def set_ticklabel_monospaced(
    fig: go.Figure, row: Union[int, None] = None, col: Union[int, None] = None
) -> None:
    """Sets the font of all ticklabels to the default monospace font.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    set_xticklabel_monospaced(fig, row, col)
    set_yticklabel_monospaced(fig, row, col)
    if _is_3d_plot(fig, row, col):
        set_zticklabel_monospaced(fig, row, col)
    if _is_plot_with_colorbar(fig):
        set_colorbar_ticklabel_monospaced(fig, row, col)


def capitalize_xticklabels(
    fig: go.Figure, row: Union[int, None] = None, col: Union[int, None] = None
) -> None:
    """Capitalizes the x-axis ticklabels.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    if _is_3d_plot(fig, row, col):
        ticktext = fig.scenes[0].xaxis_ticktext  # type: ignore
        if ticktext.islower():
            fig.update_scenes(
                xaxis_ticktext=ticktext.capitalize(),
                row=row,
                col=col,
            )
    else:
        capitalized_ticklabels = [
            label.capitalize() if label.islower() else label
            for label in fig.data[0].x  # type: ignore
        ]
        fig.update_xaxes(ticktext=capitalized_ticklabels, row=row, col=col)


def capitalize_yticklabels(
    fig: go.Figure, row: Union[int, None] = None, col: Union[int, None] = None
) -> None:
    """Capitalizes the y-axis ticklabels if all letters are lowercase.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    if _is_3d_plot(fig, row, col):
        ticktext = fig.scenes[0].yaxis_ticktext  # type: ignore
        if ticktext.islower():
            fig.update_scenes(
                yaxis_ticktext=ticktext.capitalize(),
                row=row,
                col=col,
            )
    else:
        capitalized_ticklabels = [
            label.capitalize() if label.islower() else label
            for label in fig.data[0].y  # type: ignore
        ]
        fig.update_yaxes(ticktext=capitalized_ticklabels, row=row, col=col)


def capitalize_zticklabels(
    fig: go.Figure, row: Union[int, None] = None, col: Union[int, None] = None
) -> None:
    """Capitalizes the z-axis ticklabels if all letters are lowercase.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    ticktext = fig.scenes[0].zaxis_ticktext  # type: ignore
    if ticktext.islower():
        fig.update_scenes(
            zaxis_ticktext=ticktext.capitalize(),
            row=row,
            col=col,
        )


def capitalize_ticklabels(
    fig: go.Figure, row: Union[int, None] = None, col: Union[int, None] = None
) -> None:
    """Capitalizes ticklabels for the x- and y-axes, and z-axis if applicable,
    if all letters are lowercase.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    capitalize_xticklabels(fig, row, col)
    capitalize_yticklabels(fig, row, col)
    if _is_3d_plot(fig, row, col):
        capitalize_zticklabels(fig, row, col)


def add_commas_to_xaxis_ticklabels(
    fig: go.Figure, row: Union[int, None] = None, col: Union[int, None] = None
) -> None:
    """Adds commas to the x-axis ticklabels.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    if _is_3d_plot(fig, row, col):
        fig.update_scenes(xaxis_tickformat=",", row=row, col=col)
    else:
        fig.update_xaxes(tickformat=",", row=row, col=col)


def add_commas_to_yaxis_ticklabels(
    fig: go.Figure, row: Union[int, None] = None, col: Union[int, None] = None
) -> None:
    """Adds commas to the y-axis ticklabels.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    if _is_3d_plot(fig, row, col):
        fig.update_scenes(yaxis_tickformat=",", row=row, col=col)
    else:
        fig.update_yaxes(tickformat=",", row=row, col=col)


def add_commas_to_zaxis_ticklabels(
    fig: go.Figure, row: Union[int, None] = None, col: Union[int, None] = None
) -> None:
    """Adds commas to the z-axis ticklabels.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    fig.update_scenes(zaxis_tickformat=",", row=row, col=col)


def add_commas_to_axis_tick_labels(
    fig: go.Figure, row: Union[int, None] = None, col: Union[int, None] = None
) -> None:
    """Adds commas to the numbers used for axis ticklabels.

    For example, transform 1000000 to 1,000,000.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    add_commas_to_xaxis_ticklabels(fig, row, col)
    add_commas_to_yaxis_ticklabels(fig, row, col)
    if _is_3d_plot(fig, row, col):
        add_commas_to_zaxis_ticklabels(fig, row, col)


def set_xaxis_categorical(
    fig: go.Figure, row: Union[int, None] = None, col: Union[int, None] = None
) -> None:
    """Sets the style of the x-axis to a categorical axis by removing ticks.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    # TODO: We should also adjust the margins between the ticklabels and the axis labels.
    if _is_3d_plot(fig, row, col):
        fig.update_scenes(xaxis_ticks="", row=row, col=col)
    else:
        fig.update_xaxes(ticks="", row=row, col=col)


def set_yaxis_categorical(
    fig: go.Figure, row: Union[int, None] = None, col: Union[int, None] = None
) -> None:
    """Sets the style of the y-axis to a categorical axis by removing ticks.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    # TODO: We should also adjust the margins between the ticklabels and the axis labels.
    if _is_3d_plot(fig, row, col):
        fig.update_scenes(yaxis_ticks="", row=row, col=col)
    else:
        fig.update_yaxes(ticks="", row=row, col=col)


def set_zaxis_categorical(
    fig: go.Figure, row: Union[int, None] = None, col: Union[int, None] = None
) -> None:
    """Sets the style of the z-axis to a categorical axis by removing ticks.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    fig.update_scenes(zaxis_ticks="", row=row, col=col)


def set_axes_categorical(
    fig: go.Figure, row: Union[int, None] = None, col: Union[int, None] = None
) -> None:
    """Sets all axes to categorical axes by removing ticks.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    set_xaxis_categorical(fig, row, col)
    set_yaxis_categorical(fig, row, col)
    if _is_3d_plot(fig, row, col):
        set_zaxis_categorical(fig, row, col)


def capitalize_ylabel(
    fig: go.Figure, row: Union[int, None] = None, col: Union[int, None] = None
) -> None:
    """Capitalizes the y-axis label if all letters are lowercase.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    if _is_3d_plot(fig, row, col):
        label = fig.layout.scene.yaxis.title.text  # type: ignore
        if label and label.islower():
            fig.update_scenes(
                yaxis_title_text=label.capitalize(),
                row=row,
                col=col,
            )
    else:
        label = fig.layout.yaxis.title.text  # type: ignore
        if label and label.islower():
            fig.update_yaxes(title_text=label.capitalize(), row=row, col=col)


def capitalize_xlabel(
    fig: go.Figure, row: Union[int, None] = None, col: Union[int, None] = None
) -> None:
    """Capitalizes the x-axis label if all letters are lowercase.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    if _is_3d_plot(fig, row, col):
        label = fig.layout.scene.xaxis.title.text  # type: ignore
        if label and label.islower():
            fig.update_scenes(
                xaxis_title_text=label.capitalize(),
                row=row,
                col=col,
            )
    else:
        label = fig.layout.xaxis.title.text  # type: ignore
        if label and label.islower():
            fig.update_xaxes(title_text=label.capitalize(), row=row, col=col)


def capitalize_zlabel(
    fig: go.Figure, row: Union[int, None] = None, col: Union[int, None] = None
) -> None:
    """Capitalizes the z-axis label if all letters are lowercase.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    label = fig.layout.scene.zaxis.title.text  # type: ignore
    if label and label.islower():
        fig.update_scenes(zaxis_title_text=label.capitalize(), row=row, col=col)


def capitalize_colorbar_label(
    fig: go.Figure, row: Union[int, None] = None, col: Union[int, None] = None
) -> None:
    """Capitalizes the colorbar label if all letters are lowercase.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    label = fig.layout.coloraxis.colorbar.title.text  # type: ignore
    if label and label.islower():
        new_label = label.capitalize()
        fig.update_coloraxes(colorbar_title_text=new_label, row=row, col=col)


def capitalize_axislabels(
    fig: go.Figure, row: Union[int, None] = None, col: Union[int, None] = None
) -> None:
    """Capitalizes all axis labels if all letters are lowercase.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    capitalize_xlabel(fig, row, col)
    capitalize_ylabel(fig, row, col)
    if _is_3d_plot(fig, row, col):
        capitalize_zlabel(fig, row, col)
    if _is_plot_with_colorbar(fig):
        capitalize_colorbar_label(fig, row, col)


def hide_yaxis_ticks(
    fig: go.Figure, row: Union[int, None] = None, col: Union[int, None] = None
) -> None:
    """Hides the ticks and ticklabels on the y-axis.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    fig.update_yaxes(ticks="", showticklabels=False, row=row, col=col)
    if _is_3d_plot(fig, row, col):
        fig.update_scenes(yaxis_ticks="", yaxis_showticklabels=False, row=row, col=col)


def hide_xaxis_ticks(
    fig: go.Figure, row: Union[int, None] = None, col: Union[int, None] = None
) -> None:
    """Hides the ticks and ticklabelson the x-axis.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    fig.update_xaxes(ticks="", showticklabels=False, row=row, col=col)
    if _is_3d_plot(fig, row, col):
        fig.update_scenes(xaxis_ticks="", xaxis_showticklabels=False, row=row, col=col)


def hide_zaxis_ticks(
    fig: go.Figure, row: Union[int, None] = None, col: Union[int, None] = None
) -> None:
    """Hides the ticks and ticklabels on the z-axis.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    fig.update_scenes(zaxis_ticks="", zaxis_showticklabels=False, row=row, col=col)


def hide_ticks(fig: go.Figure, row: Union[int, None] = None, col: Union[int, None] = None) -> None:
    """Hides the ticks and ticklabels on all axes.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    hide_xaxis_ticks(fig, row, col)
    hide_yaxis_ticks(fig, row, col)
    if _is_3d_plot(fig, row, col):
        hide_zaxis_ticks(fig, row, col)


def hide_yaxis_line(
    fig: go.Figure, row: Union[int, None] = None, col: Union[int, None] = None
) -> None:
    """Hides the y-axis line.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    if _is_3d_plot(fig, row, col):
        fig.update_scenes(yaxis_showline=False, row=row, col=col)
    else:
        fig.update_yaxes(showline=False, row=row, col=col)


def hide_xaxis_line(
    fig: go.Figure, row: Union[int, None] = None, col: Union[int, None] = None
) -> None:
    """Hides the x-axis line.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    if _is_3d_plot(fig, row, col):
        fig.update_scenes(xaxis_showline=False, row=row, col=col)
    else:
        fig.update_xaxes(showline=False, row=row, col=col)


def hide_zaxis_line(
    fig: go.Figure, row: Union[int, None] = None, col: Union[int, None] = None
) -> None:
    """Hides the z-axis line.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    fig.update_scenes(zaxis_showline=False, row=row, col=col)


def hide_axis_lines(
    fig: go.Figure, row: Union[int, None] = None, col: Union[int, None] = None
) -> None:
    """Hides all axis lines.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    hide_yaxis_line(fig, row, col)
    hide_xaxis_line(fig, row, col)
    if _is_3d_plot(fig, row, col):
        hide_zaxis_line(fig, row, col)


def capitalize_legend_title(fig: go.Figure) -> None:
    """Capitalizes the legend title if all letters are lowercase.

    Plotly does have the `legend_font_textcase` attribute, but the CSS styles are not
    applied correctly in SVG exports, so we manually mutate the legend entries instead.
    """
    legend_title_text = fig.layout.legend.title.text  # type: ignore
    if legend_title_text and legend_title_text.islower():
        fig.update_layout(legend_title_text=legend_title_text.capitalize())


def capitalize_legend_entries(fig: go.Figure) -> None:
    """Capitalizes the legend entries if all letters are lowercase.

    Plotly does have the `legend_font_textcase` attribute, but the CSS styles are not
    applied correctly in SVG exports, so we manually mutate the legend entries instead.
    """
    for trace in fig.data:
        if trace.name and trace.name.islower():  # type: ignore
            trace.name = trace.name.capitalize()  # type: ignore


def style_legend(fig: go.Figure) -> None:
    """Styles the legend according to Arcadia's style guide."""
    capitalize_legend_title(fig)
    capitalize_legend_entries(fig)


def get_arcadia_styles() -> dict[str, Any]:
    """Returns the Arcadia Plotly layout template as a dictionary."""
    return ARCADIA_PLOTLY_TEMPLATE_LAYOUT.to_plotly_json()


def style_plot(
    fig: go.Figure,
    monospaced_axes: Union[AxisSelector, None] = None,
    categorical_axes: Union[AxisSelector, None] = None,
    row: Union[int, None] = None,
    col: Union[int, None] = None,
) -> None:
    """Styles the plot according to Arcadia's style guide.

    Args:
        fig (go.Figure): The Plotly figure to modify.
        monospaced_axes (AxisSelector, optional): Which axes to set to the default monospaced font.
        categorical_axes (AxisSelector, optional): Which axes to set to categorical.
        row (int, optional): The row index of the subplot to modify.
        col (int, optional): The column index of the subplot to modify.
    """
    valid_axes = get_args(AxisSelector)

    if monospaced_axes is not None and monospaced_axes not in valid_axes:
        raise ValueError(f"monospaced_axes must be one of {valid_axes}, got {monospaced_axes}")
    if categorical_axes is not None and categorical_axes not in valid_axes:
        raise ValueError(f"categorical_axes must be one of {valid_axes}, got {categorical_axes}")

    capitalize_axislabels(fig, row, col)

    if categorical_axes == "all":
        set_axes_categorical(fig, row, col)
    if categorical_axes in ("x", "xy", "xz"):
        set_xaxis_categorical(fig, row, col)
    if categorical_axes in ("y", "xy", "yz"):
        set_yaxis_categorical(fig, row, col)
    if categorical_axes in ("z", "yz", "xz"):
        set_zaxis_categorical(fig, row, col)

    if monospaced_axes == "all":
        set_ticklabel_monospaced(fig, row, col)
        add_commas_to_axis_tick_labels(fig, row, col)
    if monospaced_axes in ("x", "xy", "xz"):
        set_xticklabel_monospaced(fig, row, col)
        add_commas_to_xaxis_ticklabels(fig, row, col)
    if monospaced_axes in ("y", "xy", "yz"):
        set_yticklabel_monospaced(fig, row, col)
        add_commas_to_yaxis_ticklabels(fig, row, col)
    if monospaced_axes in ("z", "yz", "xz"):
        set_zticklabel_monospaced(fig, row, col)
        add_commas_to_zaxis_ticklabels(fig, row, col)

    if _is_plot_with_legend(fig):
        style_legend(fig)

    if _is_plot_with_colorbar(fig):
        set_colorbar_ticklabel_monospaced(fig, row, col)

    # For 3D plots, we overwrite the default margin from the template.
    if _is_plot_with_3d_traces_only(fig):
        fig.update_layout(margin=dict(l=40, r=40, t=40, b=40))


def set_figure_dimensions(fig: go.Figure, size: FigureSize) -> None:
    """Sets the width and height of a figure.

    Args:
        fig (go.Figure): The figure to modify.
        size (FigureSize): The size of the figure, which must be one of the following:
            - "full_wide"
            - "full_square"
            - "float_wide"
            - "float_square"
            - "half_square"

    Raises:
        ValueError: If the size is not one of the predefined panel sizes.
    """
    if size not in FIGURE_SIZES_IN_PIXELS:
        raise ValueError(f"Size must be one of {list(FIGURE_SIZES_IN_PIXELS.keys())}.")

    width, height = FIGURE_SIZES_IN_PIXELS[size]
    fig.update_layout(width=width, height=height)


def setup() -> None:
    """Loads Arcadia fonts and styles into Plotly."""
    arcadia_template = go.layout.Template(layout=ARCADIA_PLOTLY_TEMPLATE_LAYOUT)
    pio.templates["arcadia"] = arcadia_template
    pio.templates.default = "arcadia"



import arcadia_pycolor.colors as colors
from arcadia_pycolor.gradient import Gradient

# Perceptually Uniform Gradients.
magma = Gradient(
    "magma",
    [colors.concord, colors.tanzanite, colors.heather, colors.tumbleweed, colors.wheat],
    [0, 0.217, 0.498, 0.799, 1],
)

viridis = Gradient(
    "viridis",
    [colors.space, colors.aegean, colors.lime, colors.butter],
    [0, 0.468, 0.746, 1],
)

verde = Gradient(
    "verde",
    [colors.depths, colors.shire, colors.topaz, colors.putty],
    [0, 0.357, 0.909, 1],
)

sunset = Gradient(
    "sunset",
    [colors.soil, colors.umber, colors.tumbleweed, colors.topaz, colors.putty],
    [0.0, 0.407, 0.767, 0.915, 1.0],
)

wine = Gradient(
    "wine",
    [colors.redwood, colors.dragon, colors.tangerine, colors.dawn],
    [0, 0.451, 0.828, 1],
)

lisafrank = Gradient(
    "lisafrank",
    [colors.depths, colors.aegean, colors.wish, colors.blossom],
    [0, 0.484, 0.862, 1],
)

# Strong Monocolor Gradients.
reds = Gradient(
    "reds",
    [colors.cinnabar, colors.dragon, colors.blush],
    [0.0, 0.212, 1.0],
)

oranges = Gradient(
    "oranges",
    [colors.terracotta, colors.tangerine, colors.dawn],
    [0.0, 0.761, 1.0],
)

greens = Gradient(
    "greens",
    [colors.fern, colors.lime, colors.lichen],
    [0, 0.622, 1],
)

sages = Gradient(
    "sages",
    [colors.asparagus, colors.sage, colors.lichen],
    [0, 0.641, 1],
)

blues = Gradient(
    "blues",
    [colors.lapis, colors.aegean, colors.zephyr],
    [0, 0.254, 1.0],
)

purples = Gradient(
    "purples",
    [colors.lilac, colors.aster, colors.ghost],
    [0, 0.144, 1.0],
)

# Strong Bicolor Gradients.
orange_sage = oranges + sages.reverse()
orange_sage.name = "orange_sage"

red_blue = reds + blues.reverse()
red_blue.name = "red_blue"

purple_green = purples + greens.reverse()
purple_green.name = "purple_green"

all_gradients = [obj for obj in globals().values() if isinstance(obj, Gradient)]



import pytest

from arcadia_pycolor import Gradient, HexCode, Palette


def test_palette_to_cmap():
    hex_codes = [HexCode("white", "#FFFFFF"), HexCode("black", "#000000")]
    assert Palette("my_palette", hex_codes).to_mpl_cmap().colors == hex_codes  # type: ignore


def test_palette_name():
    hex_codes = [HexCode("white", "#FFFFFF"), HexCode("black", "#000000")]
    assert (
        Palette.from_dict("my_palette", {"white": "#FFFFFF", "black": "#000000"})
        .to_mpl_cmap()
        .colors  # type: ignore
        == hex_codes
    )


@pytest.mark.parametrize(
    "values",
    [
        [0, 1],
        None,
    ],
)
def test_gradient_to_linear_cmap(values):
    grad = Gradient(
        "my_gradient",
        [HexCode("white", "#FFFFFF"), HexCode("black", "#000000")],
        values,
    ).to_mpl_cmap()
    assert grad(0) == (1.0, 1.0, 1.0, 1.0)


@pytest.mark.parametrize(
    "values",
    [
        [0, 1],
        None,
    ],
)
def test_gradient_name(values):
    name = "my_gradient"
    grad = Gradient(
        name,
        [HexCode("white", "#FFFFFF"), HexCode("black", "#000000")],
        values,
    ).to_mpl_cmap()
    assert grad.name == name



import pytest

from arcadia_pycolor import HexCode, Palette

from .test_hexcode import INVALID_HEXCODES


def test_palette_from_hexcode_list():
    hex_codes = [HexCode("white", "#FFFFFF"), HexCode("black", "#000000")]
    assert Palette("my_palette", hex_codes).colors == hex_codes


@pytest.mark.parametrize(
    "invalid_input",
    INVALID_HEXCODES,
)
def test_palette_from_hexcode_list_invalid_input(invalid_input):
    with pytest.raises(ValueError):
        Palette("a", [invalid_input])


def test_palette_from_dict():
    hex_codes = [HexCode("white", "#FFFFFF"), HexCode("black", "#000000")]
    assert (
        Palette.from_dict(
            "my_palette",
            {"white": "#FFFFFF", "black": "#000000"},
        ).colors
        == hex_codes
    )


def test_palette_repr():
    expected_swatch = "\x1b[48;2;255;255;255m  \x1b[0m\x1b[48;2;0;0;0m  \x1b[0m\n\x1b[48;2;255;255;255m  \x1b[0m\x1b[38;2;255;255;255m white #FFFFFF\x1b[0m\n\x1b[48;2;0;0;0m  \x1b[0m\x1b[38;2;0;0;0m black #000000\x1b[0m"  # noqa E501
    assert (
        Palette(
            "my_palette",
            [HexCode("white", "#FFFFFF"), HexCode("black", "#000000")],
        ).__repr__()
        == expected_swatch
    )


def test_palette_add():
    p1 = [HexCode("white", "#FFFFFF")]
    p2 = [HexCode("black", "#000000")]
    result = p1 + p2
    assert (Palette("p1", p1) + Palette("p2", p2)).colors == result
    assert (Palette("p1", p1) + Palette("p2", p2)).name == "p1+p2"


def test_palette_length():
    """Test that a Palette has the expected length."""
    palette = Palette("test_palette", [HexCode("white", "#FFFFFF"), HexCode("black", "#000000")])
    assert len(palette) == 2


def test_palette_iteration():
    """Test that a Palette can be iterated over."""
    colors = [HexCode("white", "#FFFFFF"), HexCode("black", "#000000")]
    palette = Palette("test_palette", colors)

    for i, color in enumerate(palette):
        assert color == colors[i]

    assert [color for color in palette] == ["#FFFFFF", "#000000"]


def test_palette_indexing():
    """Test that a Palette can be indexed."""
    colors = [HexCode("white", "#FFFFFF"), HexCode("black", "#000000")]
    palette = Palette("test_palette", colors)

    assert palette[0] == colors[0]
    assert palette[1] == colors[1]

    assert palette[-1] == colors[-1]

    with pytest.raises(IndexError):
        palette[2]


def test_palette_slicing():
    """Test that a Palette can be sliced."""
    colors = [HexCode("white", "#FFFFFF"), HexCode("gray", "#CCCCCC"), HexCode("black", "#000000")]
    palette = Palette("test_palette", colors)

    sliced = palette[0:2]
    assert isinstance(sliced, Palette)
    assert sliced.name == "test_palette_slice"
    assert len(sliced) == 2
    assert sliced.colors == colors[0:2]

    sliced = palette[-2:]
    assert len(sliced) == 2
    assert sliced.colors == colors[-2:]

    sliced = palette[::2]
    assert len(sliced) == 2
    assert sliced.colors == [colors[0], colors[2]]

    palette = Palette("test", colors)
    assert palette.reverse().colors == palette[::-1].colors



import plotly.express as px
import plotly.graph_objects as go
import pytest

import arcadia_pycolor as apc

IRIS_DATASET = px.data.iris()


def plot_3d_scatter_plot_with_plotly_express():
    """Plot a 3D scatter plot using Plotly Express."""
    fig = px.scatter_3d(
        IRIS_DATASET,
        x="sepal_length",
        y="sepal_width",
        z="petal_width",
        color="species",
        size="sepal_length",
        color_discrete_map={
            "setosa": apc.aegean,
            "versicolor": apc.amber,
            "virginica": apc.seaweed,
        },
    )
    apc.plotly.style_plot(fig, monospaced_axes="all")
    return fig


def plot_3d_scatter_plot_with_plotly_graph_objects():
    """Plot a 3D scatter plot using Plotly Graph Objects."""
    fig = go.Figure()

    fig.add_trace(
        go.Scatter3d(
            x=IRIS_DATASET["sepal_length"],
            y=IRIS_DATASET["sepal_width"],
            z=IRIS_DATASET["petal_width"],
            mode="markers",
            marker=dict(
                size=5,
                color=IRIS_DATASET["species"].map(
                    {
                        "setosa": apc.aegean,
                        "versicolor": apc.amber,
                        "virginica": apc.seaweed,
                    }
                ),
                opacity=0.8,
            ),
            text=IRIS_DATASET["species"],
            hoverinfo="text",
        )
    )

    fig.update_layout(
        scene=dict(
            xaxis_title="sepal_length",
            yaxis_title="sepal_width",
            zaxis_title="petal_width",
        ),
        showlegend=False,
    )

    apc.plotly.style_plot(fig, monospaced_axes="all")
    return fig


@pytest.mark.parametrize(
    "plotting_function",
    [
        plot_3d_scatter_plot_with_plotly_express,
        plot_3d_scatter_plot_with_plotly_graph_objects,
    ],
)
def test_3d_plots_with_plotly(output_dirpath, plotting_function):
    fig = plotting_function()
    fig.write_html(output_dirpath / f"{plotting_function.__name__}.html")



from arcadia_pycolor import colors, cvd, gradients, mpl, palettes, plot, style_defaults
from arcadia_pycolor import plotly_utils as plotly

from .colors import *
from .gradient import Gradient
from .hexcode import HexCode
from .palette import Palette

# This is a placeholder that will be replaced by the version number at build time.
__version__ = "0.7.1"

__all__ = [
    "cvd",
    "Gradient",
    "gradients",
    "HexCode",
    "mpl",
    "Palette",
    "palettes",
    "plot",
    "plotly",
    "style_defaults",
]

colors_all = [name for name in dir(colors) if isinstance(getattr(colors, name), HexCode)]

__all__.extend(colors_all)  # type: ignore
